import base64
import json
import logging
import uuid
from uuid import UUID
from datetime import datetime, timedelta
from typing import Tuple, Optional, Union

from sqlalchemy.ext.asyncio import AsyncSession
from starlette import status

from starlette_web.common.conf import settings
from starlette_web.common.email import send_email, EmailSenderError
from starlette_web.common.http.base_endpoint import BaseHTTPEndpoint
from starlette_web.common.http.exceptions import (
    AuthenticationFailedError,
    InvalidParameterError,
    SendRequestError,
    ConflictError,
    UnprocessableEntityError,
)
from starlette_web.common.utils import get_random_string
from starlette_web.contrib.auth.models import User, UserSession, UserInvite
from starlette_web.contrib.auth.backend import JWTAuthenticationBackend
from starlette_web.contrib.auth.permissions import IsSuperuserPermission
from starlette_web.contrib.auth.utils import (
    encode_jwt,
    TokenCollection,
    TOKEN_TYPE_REFRESH,
    TOKEN_TYPE_RESET_PASSWORD,
)
from starlette_web.contrib.auth.schemas import (
    SignInSchema,
    SignUpSchema,
    JWTResponseSchema,
    RefreshTokenSchema,
    UserResponseSchema,
    ChangePasswordSchema,
    UserInviteRequestSchema,
    UserInviteResponseSchema,
    ResetPasswordRequestSchema,
    ResetPasswordResponseSchema,
)
from starlette_web.contrib.auth.password_validation import validate_password

logger = logging.getLogger(__name__)


# TODO: rewrite as separate manager, not bound to views
class JWTSessionMixin:
    """Allows updating session and prepare usual / refresh JWT tokens"""

    response_schema = JWTResponseSchema
    db_session: AsyncSession = NotImplemented

    @staticmethod
    def _get_tokens(user: User, session_id: Union[str, UUID]) -> TokenCollection:
        token_payload = {"user_id": user.id, "session_id": str(session_id)}
        access_token, access_token_expired_at = encode_jwt(token_payload)
        refresh_token, refresh_token_expired_at = encode_jwt(
            token_payload,
            token_type=TOKEN_TYPE_REFRESH,
        )
        return TokenCollection(
            refresh_token=refresh_token,
            refresh_token_expired_at=refresh_token_expired_at,
            access_token=access_token,
            access_token_expired_at=access_token_expired_at,
        )

    async def _create_session(self, user: User) -> TokenCollection:
        session_id = uuid.uuid4()
        token_collection = self._get_tokens(user, session_id)
        await UserSession.async_create(
            self.db_session,
            user_id=user.id,
            public_id=str(session_id),
            refresh_token=token_collection.refresh_token,
            expired_at=token_collection.refresh_token_expired_at,
        )
        return token_collection

    async def _update_session(self, user: User, session_id: Union[str, UUID]) -> TokenCollection:
        user_session = await UserSession.async_get(self.db_session, public_id=str(session_id))
        if not user_session:
            return await self._create_session(user)

        token_collection = self._get_tokens(user, session_id=session_id)
        await user_session.update(
            self.db_session,
            refresh_token=token_collection.refresh_token,
            expired_at=token_collection.refresh_token_expired_at,
            refreshed_at=datetime.utcnow(),
            is_active=True,
        )
        return token_collection


class SignInAPIView(JWTSessionMixin, BaseHTTPEndpoint):
    """Allows to Log-In user and update/create his session"""

    request_schema = SignInSchema
    auth_backend = None

    async def post(self, request):
        """
        description: Sign in
        requestBody:
          required: true
          description: Sign in
          content:
            application/json:
              schema: SignInSchema
        responses:
          200:
            description: JsonWebToken
            content:
              application/json:
                schema: JWTResponseSchema
          401:
            description: Authentication failed
            content:
              application/json:
                schema: Error
        tags: ["Authorization"]
        """
        cleaned_data = await self._validate(request)
        user = await self.authenticate(cleaned_data["email"], cleaned_data["password"])
        token_collection = await self._create_session(user)
        return self._response(token_collection)

    async def authenticate(self, email: str, password: str) -> User:
        user = await User.async_get(db_session=self.db_session, email=email, is_active__is=True)
        if not user:
            logger.info("Not found active user with email [%s]", email)
            raise AuthenticationFailedError(
                details="Not found active user with provided email.",
            )

        if not user.verify_password(password):
            logger.error("Password didn't verify: email: %s", email)
            raise AuthenticationFailedError(
                details="Email or password is invalid.",
            )

        return user


class SignUpAPIView(JWTSessionMixin, BaseHTTPEndpoint):
    """Allows creating new user and create his own session"""

    request_schema = SignUpSchema
    auth_backend = None

    async def post(self, request):
        """
        description: Sign up
        requestBody:
          required: true
          description: Sign up
          content:
            application/json:
              schema: SignUpSchema
        responses:
          200:
            description: JsonWebToken
            content:
              application/json:
                schema: JWTResponseSchema
        tags: ["Authorization"]
        """
        cleaned_data = await self._validate(request)
        user_invite: UserInvite = cleaned_data["user_invite"]
        user = await User.async_create(
            self.db_session,
            email=cleaned_data["email"],
            password=User.make_password(cleaned_data["password_1"]),
        )
        await user_invite.update(self.db_session, is_applied=True, user_id=user.id)
        token_collection = await self._create_session(user)
        return self._response(token_collection, status_code=status.HTTP_201_CREATED)

    async def _validate(self, request, *_) -> dict:
        cleaned_data = (await super()._validate(request)) or dict()
        email = cleaned_data["email"]

        if await User.async_get(self.db_session, email=email):
            raise ConflictError(details=f"User with email '{email}' already exists")

        user_invite = await UserInvite.async_get(
            self.db_session,
            token=cleaned_data["invite_token"],
            is_applied__is=False,
            expired_at__gt=datetime.utcnow(),
        )

        if not user_invite:
            details = "Invitation link is expired or unavailable"
            logger.error(
                "Couldn't signup user token: %s | details: %s",
                cleaned_data["invite_token"],
                details,
            )
            raise UnprocessableEntityError(details=details)

        if email != user_invite.email:
            raise InvalidParameterError(message="Email does not match with your invitation.")

        validate_password(cleaned_data["password_1"], User(email=email))

        cleaned_data["user_invite"] = user_invite
        return cleaned_data


class SignOutAPIView(BaseHTTPEndpoint):
    """
    Sign-out consists from 2 operations:
     - remove JWT token on front-end side
     - deactivate current session on BE (this allows to block use regular or refresh token)
    """

    auth_backend = JWTAuthenticationBackend

    async def delete(self, request):
        """
        description: Sign out
        responses:
          200:
            description: Signed out
        tags: ["Authorization"]
        """
        user = request.user
        logger.info("Log out for user %s", user)

        user_session = await UserSession.async_get(
            self.db_session,
            public_id=self.scope["user_session_id"],
            is_active=True,
        )
        if user_session:
            logger.info("Session %s exists and active. It will be updated.", user_session)
            user_session.is_active = False
            await self.db_session.flush()

        else:
            logger.info("Not found active sessions for user %s. Skip sign-out.", user)

        return self._response(status_code=status.HTTP_200_OK)


class RefreshTokenAPIView(JWTSessionMixin, BaseHTTPEndpoint):
    """Allows updating tokens (should be called when main token is outdated)"""

    request_schema = RefreshTokenSchema
    auth_backend = None

    async def post(self, request):
        """
        description: Update refresh token
        requestBody:
          required: true
          description: Refresh token
          content:
            application/json:
              schema: RefreshTokenSchema
        responses:
          200:
            description: JsonWebToken
            content:
              application/json:
                schema: JWTResponseSchema
          401:
            description: Authentication failed
            content:
              application/json:
                schema: Error
        tags: ["Authorization"]
        """
        user, refresh_token, session_id = await self._validate(request)
        if session_id is None:
            raise AuthenticationFailedError("No session ID in token found")

        user_session = await UserSession.async_get(
            self.db_session, public_id=session_id, is_active=True
        )
        if not user_session:
            raise AuthenticationFailedError("There is not active session for user.")

        if user_session.refresh_token != refresh_token:
            raise AuthenticationFailedError("Refresh token does not match with user session.")

        token_collection = await self._update_session(user, session_id)
        return self._response(token_collection)

    async def _validate(self, request, *args, **kwargs) -> Tuple[User, str, Optional[str]]:
        cleaned_data = await super()._validate(request)
        refresh_token = cleaned_data["refresh_token"]
        user, jwt_payload, _ = await JWTAuthenticationBackend(
            request,
            self.scope,
        ).authenticate_user(refresh_token, token_type="refresh")
        return user, refresh_token, jwt_payload.get("session_id")


class InviteUserAPIView(BaseHTTPEndpoint):
    """Invite user (by email) to the service"""

    request_schema = UserInviteRequestSchema
    response_schema = UserInviteResponseSchema
    auth_backend = JWTAuthenticationBackend

    async def post(self, request):
        """
        description: Invite user
        requestBody:
          required: true
          description: Invited user's credentials
          content:
            application/json:
              schema: UserInviteRequestSchema
        responses:
          200:
            description: Invited user
            content:
              application/json:
                schema: UserInviteResponseSchema
        tags: ["Authorization"]
        """
        cleaned_data = await self._validate(request)
        email = cleaned_data["email"]
        token = UserInvite.generate_token()
        expired_at = datetime.utcnow() + timedelta(
            seconds=settings.AUTH_INVITE_LINK_EXPIRES_IN)

        if user_invite := await UserInvite.async_get(self.db_session, email=email):
            logger.info("INVITE: update for %s (expired %s) token [%s]", email, expired_at, token)
            await user_invite.update(self.db_session, token=token, expired_at=expired_at)

        else:
            logger.info("INVITE: create for %s (expired %s) token [%s]", email, expired_at, token)
            user_invite = await UserInvite.async_create(
                self.db_session,
                email=email,
                token=token,
                expired_at=expired_at,
                owner_id=request.user.id,
            )

        logger.info("Invite object %r created. Sending message...", user_invite)
        await self._send_email(user_invite)
        return self._response(user_invite, status_code=status.HTTP_201_CREATED)

    @staticmethod
    async def _send_email(user_invite: UserInvite):
        invite_data = {"token": user_invite.token, "email": user_invite.email}
        invite_data = base64.urlsafe_b64encode(json.dumps(invite_data).encode()).decode()
        link = f"{settings.SITE_URL}/sign-up/?i={invite_data}"
        body = (
            f"<p>Hello! :) You have been invited to {settings.SITE_URL}</p>"
            f"<p>Please follow the link </p>"
            f"<p><a href={link}>{link}</a></p>"
        )

        try:
            await send_email(
                recipients_list=[user_invite.email],
                subject=f"Welcome to {settings.SITE_URL}",
                html_content=body.strip(),
            )
        except EmailSenderError as exc:
            raise SendRequestError(**dict(exc))

    async def _validate(self, request, *_) -> dict:
        cleaned_data = (await super()._validate(request)) or dict()

        if exists_user := await User.async_get(self.db_session, email=cleaned_data["email"]):
            raise InvalidParameterError(f"User with email=[{exists_user.email}] already exists.")

        return cleaned_data


class ResetPasswordAPIView(BaseHTTPEndpoint):
    """Send link to user's email for resetting his password"""

    request_schema = ResetPasswordRequestSchema
    response_schema = ResetPasswordResponseSchema
    auth_backend = JWTAuthenticationBackend
    permission_classes = [IsSuperuserPermission]

    async def post(self, request):
        """
        description: Reset password
        requestBody:
          required: true
          description: User credentials
          content:
            application/json:
              schema: ResetPasswordRequestSchema
        responses:
          200:
            description: Link to email
            content:
              application/json:
                schema: ResetPasswordResponseSchema
        tags: ["Authorization"]
        """
        user = await self._validate(request)
        token = self._generate_token(user)
        await self._send_email(user, token)
        return self._response({"user_id": user.id, "email": user.email, "token": token})

    async def _validate(self, request, *_) -> User:
        cleaned_data = await super()._validate(request)
        user = await User.async_get(self.db_session, email=cleaned_data["email"])
        if not user:
            raise InvalidParameterError(f"User with email=[{cleaned_data['email']}] not found.")

        return user

    @staticmethod
    async def _send_email(user: User, token: str):
        link = f"{settings.SITE_URL}/change-password/?t={token}"
        body = (
            f"<p>You can reset your password for {settings.SITE_URL}</p>"
            f"<p>Please follow the link </p>"
            f"<p><a href={link}>{link}</a></p>"
        )
        try:
            await send_email(
                recipients_list=[user.email],
                subject=f"Welcome back to {settings.SITE_URL}",
                html_content=body.strip(),
            )
        except EmailSenderError as exc:
            raise SendRequestError(**dict(exc))

    @staticmethod
    def _generate_token(user: User) -> str:
        payload = {
            "user_id": user.id,
            "email": user.email,
            "jti": f"token-{uuid.uuid4()}",  # JWT ID
            "slt": get_random_string(length=12),
        }
        token, _ = encode_jwt(
            payload,
            token_type=TOKEN_TYPE_RESET_PASSWORD,
            expires_in=settings.AUTH_RESET_PASSWORD_LINK_EXPIRES_IN,
        )
        return token


class ChangePasswordAPIView(JWTSessionMixin, BaseHTTPEndpoint):
    """Simple API for changing user's password"""

    request_schema = ChangePasswordSchema
    auth_backend = None

    async def post(self, request):
        """
        description: Change password
        requestBody:
          required: true
          description: New password confirmation
          content:
            application/json:
              schema: ChangePasswordSchema
        responses:
          200:
            description: New tokens
            content:
              application/json:
                schema: JWTResponseSchema
          401:
            description: Authentication failed
        tags: ["Authorization"]
        """
        cleaned_data = await self._validate(request)
        validate_password(cleaned_data["password_1"], None)

        user, _, _ = await JWTAuthenticationBackend(request, self.scope).authenticate_user(
            jwt_token=cleaned_data["token"],
            token_type=TOKEN_TYPE_RESET_PASSWORD,
        )
        new_password = User.make_password(cleaned_data["password_1"])
        await user.update(self.db_session, password=new_password)

        token_collection = await self._create_session(user)
        return self._response(token_collection)


class ProfileApiView(BaseHTTPEndpoint):
    """Simple retrieves profile information (for authenticated user)"""

    response_schema = UserResponseSchema
    auth_backend = JWTAuthenticationBackend

    async def get(self, request):
        """
        description: Profile info
        responses:
          200:
            description: Profile
            content:
              application/json:
                schema: UserResponseSchema
        tags: ["Authorization"]
        """
        return self._response(request.user)



from marshmallow.validate import ValidationError, Email as EmailValidator

from starlette_web.common.http.exceptions import InvalidParameterError
from starlette_web.common.management.base import CommandError
from starlette_web.contrib.auth.password_validation import validate_password as _validate_password


class AuthCommandMixin:
    # TODO: Support masking for passwords ?
    def get_input_data(self, message, default=None):
        raw_value = input(message)
        if default and raw_value == "":
            raw_value = default
        return raw_value

    def validate_field(self, field_name: str, value: str):
        if not value:
            raise CommandError(details="Field value is empty.")

        if field_name == "email":
            try:
                EmailValidator()(value)
            except ValidationError:
                raise CommandError(details="Invalid value for email.")

    def validate_password(self, password, user=None):
        try:
            _validate_password(password, user=user)
        except InvalidParameterError as exc:
            print(f"Weak password: {exc.details}")
            answer = input("Skip validation and continue? (y/n): ")

            if answer.strip() not in ["y", "Y", "yes"]:
                raise exc



# flake8: noqa

import os


os.environ.setdefault("STARLETTE_SETTINGS_MODULE", "core.settings")
from starlette_web.common.app import get_app
app = get_app()


if __name__ == "__main__":
    import uvicorn
    import sys

    for arg in sys.argv:
        if arg.startswith("--settings="):
            settings_module = arg[11:]
            os.environ["STARLETTE_SETTINGS_MODULE"] = settings_module

    uvicorn.run(app, host="127.0.0.1", port=80)



from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Tuple

import jwt

from starlette_web.common.conf import settings


TOKEN_TYPE_ACCESS = "access"
TOKEN_TYPE_REFRESH = "refresh"
TOKEN_TYPE_RESET_PASSWORD = "reset_password"


@dataclass
class TokenCollection:
    refresh_token: str
    refresh_token_expired_at: datetime
    access_token: str
    access_token_expired_at: datetime


def encode_jwt(
    payload: dict,
    token_type: str = TOKEN_TYPE_ACCESS,
    expires_in: int = None,
) -> Tuple[str, datetime]:
    """Allows to prepare JWT for auth engine"""

    if token_type == TOKEN_TYPE_REFRESH:
        expires_in = settings.AUTH_JWT_REFRESH_EXPIRES_IN
    else:
        expires_in = expires_in or settings.AUTH_JWT_EXPIRES_IN

    expired_at = datetime.utcnow() + timedelta(seconds=expires_in)
    payload["exp"] = expired_at
    payload["exp_iso"] = expired_at.isoformat()
    payload["token_type"] = token_type
    token = jwt.encode(
        payload,
        str(settings.SECRET_KEY),
        algorithm=settings.AUTH_JWT_ALGORITHM,
    )
    return token, expired_at


def decode_jwt(encoded_jwt: str) -> dict:
    """Allows to decode received JWT token to payload"""

    return jwt.decode(
        encoded_jwt,
        str(settings.SECRET_KEY),
        algorithms=[settings.AUTH_JWT_ALGORITHM],
    )



import json
import hashlib
import logging
import os
import tempfile
from typing import List, Dict, Tuple, Union

import anyio
from filelock import FileLock
from traceback_with_variables import format_exc

from starlette_web.common.conf import settings as project_settings

# Scheduler is only supposed to be run as management-command
from starlette_web.common.management.base import CommandError
from starlette_web.common.utils.importing import import_string
from starlette_web.contrib.scheduler.app_settings import Settings


logger = logging.getLogger("starlette_web.contrib.scheduler")
JSONType = Union[
    Dict[str, "JSONType"],
    List["JSONType"],
    str,
    int,
    float,
    bool,
    None,
]
# Arguments:
# [0] crontab-like schedule (including @reboot)
# [1] import string for async function to be executed
# [2] args
# [3] kwargs
# [4] timeout
JobType = Tuple[str, str, List[JSONType], Dict[str, JSONType], Union[float, None]]


class BasePeriodicTaskScheduler:
    settings_class = Settings

    def __init__(self, **kwargs):
        self.needs_read = kwargs.get("needs_read", True)
        self.needs_write = kwargs.get("needs_write", True)
        self.settings = self.settings_class(project_settings)

    def __enter__(self):
        if self.needs_read:
            self._read_jobs()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.needs_write:
            self._write_jobs()
        return False

    def _read_jobs(self):
        raise CommandError("Not supported for this platform.")

    def _write_jobs(self):
        raise CommandError("Not supported for this platform.")

    def add_jobs(self):
        raise CommandError("Not supported for this platform.")

    def update_jobs(self):
        self.remove_jobs()
        self.add_jobs()

    def show_jobs(self):
        raise CommandError("Not supported for this platform.")

    def remove_jobs(self):
        raise CommandError("Not supported for this platform.")

    async def run_job(self, job_hash: str):
        cron_pattern, async_func, job_args, job_kwargs, timeout = self._get_job_by_hash(job_hash)
        async_job_handler = import_string(async_func)

        with self._get_job_mutex(job_hash):
            with anyio.CancelScope() as cancel_scope:
                if timeout is not None:
                    cancel_scope.deadline = anyio.current_time() + timeout

                try:
                    await async_job_handler(*job_args, **job_kwargs)
                except Exception as exc:
                    logger.critical(format_exc(exc))

    def _get_project_level_hash(self) -> str:
        return hashlib.md5(str(project_settings.SECRET_KEY).encode("utf-8")).hexdigest()

    def _hash_job(self, job: JobType) -> str:
        job_string = json.JSONEncoder(sort_keys=True).encode(job)
        job_hash = hashlib.md5(job_string.encode("utf-8")).hexdigest()
        return job_hash

    def _get_job_by_hash(self, job_hash: str) -> JobType:
        for job in self.settings.PERIODIC_JOBS:
            if self._hash_job(job) == job_hash:
                return job

        raise CommandError(details=f"Periodic job with hash = {job_hash} not found.")

    def _get_job_mutex(self, job_hash: str):
        if self.settings.LOCK_JOBS:
            lock_file = os.path.join(
                tempfile.gettempdir(), self._get_project_level_hash() + "_" + job_hash + ".lock"
            )
            return FileLock(lock_file, timeout=self.settings.BLOCKING_TIMEOUT)
        else:

            class DummyLock:
                def __enter__(self):
                    return self

                def __exit__(self, exc_type, exc_val, exc_tb):
                    return False

            return DummyLock()



import os
import shutil
from pathlib import Path

from starlette_web.common.management.alembic_mixin import AlembicMixin
from starlette_web.common.management.base import BaseCommand, CommandError, CommandParser
from starlette_web.common.utils import get_random_string


class Command(BaseCommand, AlembicMixin):
    help = "Initialize directory with project files"
    _alembic_directory_name = "alembic"

    def add_arguments(self, parser: CommandParser):
        parser.add_argument("project_name", type=str)

    async def handle(self, **options):
        current_dir = os.getcwd()
        project_name = options["project_name"]

        cwd = Path(current_dir)
        project_dir = cwd / project_name
        if project_dir.is_file() or project_dir.is_symlink():
            raise CommandError(
                details=(
                    f"Cannot create project directory {project_name}. "
                    "A file/link with such name exists in the current directory."
                )
            )

        if project_dir.is_dir():
            raise CommandError(
                details=f"Directory {project_dir} already exists. Exiting."
            )

        project_dir.mkdir()
        defaults_dir = Path(__file__).parent / "_project_defaults"

        shutil.copytree(
            defaults_dir / "core",
            project_dir / "core",
        )
        for filename in ["command.py", "asgi.py", "__init__.py"]:
            shutil.copy(
                defaults_dir / filename,
                project_dir / filename,
            )

        env_template = """
APP_DEBUG=false
SITE_URL="https://web.project.com/"
SECRET_KEY={secret_key}

DB_HOST=localhost
DB_PORT=5432
DB_NAME=web_project
DB_USERNAME=postgres
DB_PASSWORD=postgres
DB_ECHO=false
        """

        with open(project_dir / ".env", "wt+", encoding="utf-8") as file:
            content = env_template.format(
                secret_key=get_random_string(50),
            )
            file.writelines(content.strip() + "\n")

        with open(project_dir / ".env.template", "wt+", encoding="utf-8") as file:
            content = env_template.format(
                secret_key="",
            )
            file.writelines(content.strip() + "\n")

        # Setup base directories

        (project_dir / "static").mkdir()
        (project_dir / "templates").mkdir()

        # Setup alembic
        os.chdir(project_dir)
        await self.run_alembic_main(["init", "-t", "async", self._alembic_directory_name])
        with open(project_dir / self._alembic_directory_name / "env.py", "rt") as file:
            lines = []
            for line in file:
                if line.strip() == "target_metadata = None":
                    lines += [
                        "from starlette_web.common.conf import settings\n",
                        "from starlette_web.common.conf.app_manager import app_manager\n",
                        "from starlette_web.common.database.model_base import ModelBase\n",
                        "app_manager.import_models()\n"
                        "target_metadata = ModelBase.metadata\n",
                    ]
                else:
                    lines.append(line)

        with open(project_dir / self._alembic_directory_name / "env.py", "wt") as file:
            file.writelines(lines)

        with open(project_dir / "alembic.ini", "rt") as file:
            lines = []
            for line in file:
                if "# file_template = " in line:
                    lines.append(line[2:])
                else:
                    lines.append(line)

        with open(project_dir / "alembic.ini", "wt") as file:
            file.writelines(lines)



from collections import OrderedDict
from importlib import import_module

from starlette_web.common.conf import settings
from starlette_web.common.conf.base_app_config import BaseAppConfig
from starlette_web.common.http.exceptions import ImproperlyConfigured


class AppManager:
    """
    A project-wide manager for installed modules (apps, in django terminology).
    Dynamically introspects %module_name%/apps.py and runs apps' own
    initializations and checks. Called at Starlette app initialization.
    """

    def __init__(self):
        self.app_configs: OrderedDict[str, BaseAppConfig] = OrderedDict()
        self.app_names: OrderedDict[str, str] = OrderedDict()
        self._apps_registered = False

    def register_apps(self):
        if self._apps_registered:
            return

        for installed_app in settings.INSTALLED_APPS:
            _module = import_module(installed_app)
            if not _module.__file__:
                raise ImproperlyConfigured(
                    details=f"App {installed_app} must contain an __init__.py file."
                )

            try:
                _app_module = import_module(installed_app + ".apps")
            except (SystemError, ImportError) as exc:
                raise ImproperlyConfigured(details=str(exc)) from exc

            try:
                AppConfig = getattr(_app_module, "AppConfig")
                if BaseAppConfig not in AppConfig.__mro__:
                    raise AssertionError
            except (AttributeError, AssertionError) as exc:
                raise ImproperlyConfigured(
                    details=f"App {installed_app} must define apps.AppConfig class, inherited "
                    f"from starlette_web.common.conf.base_app_config.BaseAppConfig"
                ) from exc

            app_config: BaseAppConfig = AppConfig()
            if app_config.app_name in self.app_names.keys():
                raise ImproperlyConfigured(
                    details=f"Duplicate app name {app_config.app_name}: "
                    f"{installed_app} and {self.app_names[app_config.app_name]}"
                )

            for requirement in app_config.app_requirements:
                if requirement not in self.app_names.values():
                    raise ImproperlyConfigured(
                        details=f"App {installed_app} requires {requirement} "
                        f"to be defined in settings.INSTALLED_APPS strictly before it."
                    )

            self.app_names[app_config.app_name] = installed_app
            self.app_configs[installed_app] = app_config

        self._apps_registered = True

    def initialize_apps(self):
        self.register_apps()
        for app_config in self.app_configs.values():
            app_config.initialize()

    def run_apps_checks(self):
        self.register_apps()
        for app_config in self.app_configs.values():
            app_config.perform_checks()

    def import_models(self):
        self.register_apps()
        for installed_app in settings.INSTALLED_APPS:
            try:
                import_module(installed_app + ".models")
            except (SystemError, ImportError):
                pass


app_manager = AppManager()



import pytest

from starlette_web.common.conf import settings
from starlette_web.common.http.exceptions import ImproperlyConfigured


def test_settings():
    value = settings.APPLICATION_CLASS
    assert value == "starlette_web.common.app.BaseStarletteApplication"

    with pytest.raises(ImproperlyConfigured) as err:
        _ = settings.INEXISTENT_SETTING

    assert err.value.details == "Setting INEXISTENT_SETTING is not configured."
    assert err.value.message == "Application is configured improperly."



from typing import Type

from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import NullPool

from starlette_web.common.conf import settings
from starlette_web.common.utils import import_string


def get_async_session_class() -> Type[AsyncSession]:
    return import_string(settings.DB_ASYNC_SESSION_CLASS)


def make_session_maker(**kwargs) -> sessionmaker:
    use_pool = kwargs.get("use_pool", True)
    connect_args = kwargs.get("connect_args", {"timeout": 20})

    if use_pool:
        pool_size = settings.DATABASE["pool_min_size"]
        max_overflow = max(0, settings.DATABASE["pool_max_size"] - pool_size)

        # If poolclass is None, uses one implemented by engine's method .get_pool_class()
        # Namely, for asyncpg, sqlalchemy.pool.impl.AsyncAdaptedQueuePool
        create_async_engine_kw = dict(
            poolclass=None,
            pool_size=pool_size,
            max_overflow=max_overflow,
            pool_recycle=settings.DB_POOL_RECYCLE,
        )
    else:
        create_async_engine_kw = dict(
            poolclass=NullPool,
        )

    db_engine = create_async_engine(
        settings.DATABASE_DSN,
        echo=settings.DB_ECHO,
        connect_args=connect_args,
        **create_async_engine_kw,
    )

    return sessionmaker(
        db_engine,
        expire_on_commit=False,
        class_=get_async_session_class(),
        autoflush=True,
        autocommit=False,
    )



from starlette.routing import Mount, Route, WebSocketRoute
from starlette.staticfiles import StaticFiles

from starlette_web.common.conf import settings
from starlette_web.contrib.apispec.views import OpenApiView
from starlette_web.contrib.auth.routes import routes as auth_routes
from starlette_web.contrib.admin import admin, AdminMount
from starlette_web.tests.views import (
    HealthCheckAPIView,
    BaseWebsocketTestEndpoint,
    CancellationWebsocketTestEndpoint,
    AuthenticationWebsocketTestEndpoint,
    FinitePeriodicTaskWebsocketTestEndpoint,
    InfinitePeriodicTaskWebsocketTestEndpoint,
    ChatWebsocketTestEndpoint,
)


# TODO: split auth and api
routes = [
    Mount("/api", routes=auth_routes),
    Route("/openapi", OpenApiView, include_in_schema=False),
    AdminMount("/admin", app=admin.get_app(), name="admin"),
    Mount("/static", app=StaticFiles(directory=settings.STATIC["ROOT_DIR"]), name="static"),
    Mount("/media", app=StaticFiles(directory=settings.MEDIA["ROOT_DIR"]), name="media"),
    Route("/health_check/", HealthCheckAPIView),
    WebSocketRoute("/ws/test_websocket_base", BaseWebsocketTestEndpoint),
    WebSocketRoute("/ws/test_websocket_cancel", CancellationWebsocketTestEndpoint),
    WebSocketRoute("/ws/test_websocket_auth", AuthenticationWebsocketTestEndpoint),
    WebSocketRoute("/ws/test_websocket_finite_periodic", FinitePeriodicTaskWebsocketTestEndpoint),
    WebSocketRoute(
        "/ws/test_websocket_infinite_periodic", InfinitePeriodicTaskWebsocketTestEndpoint
    ),
    WebSocketRoute("/ws/test_websocket_chat", ChatWebsocketTestEndpoint),
]



from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Optional, List

import aiosmtplib
import anyio

from starlette_web.common.email.base_sender import BaseEmailSender
from starlette_web.common.http.exceptions import NotSupportedError
from starlette_web.common.utils import get_available_options


class SMTPEmailSender(BaseEmailSender):
    SEND_MAX_TIME = 30

    async def send_email(
        self,
        subject: str,
        html_content: str,
        recipients_list: List[str],
        from_email: Optional[str] = None,
    ):
        message = MIMEMultipart("alternative")
        message["Subject"] = subject
        _from = from_email
        if _from is None:
            _from = self.options.get("from")
        if _from is None:
            raise NotSupportedError(
                details="Cannot send email without setting FROM field."
            )
        message["From"] = _from

        html_part = MIMEText(html_content, "html")
        message.attach(html_part)

        _recipients = recipients_list.copy()
        while _recipients:
            available_args = get_available_options(aiosmtplib.send)
            options = dict(
                message=message,
                sender=_from,
                recipients=_recipients[:self.MAX_BULK_SIZE],
                **self.options,
            )

            with anyio.move_on_after(self.SEND_MAX_TIME):
                await aiosmtplib.send(**{
                    key: value
                    for key, value in options.items()
                    if key in available_args
                })

            del _recipients[:self.MAX_BULK_SIZE]



import sys
from typing import IO, Any, AnyStr, List, Literal, Optional
from typing import AsyncContextManager, AsyncIterator

import anyio

from starlette_web.common.http.exceptions import NotSupportedError


MODE = Literal["t", "b"]


class DummyAsyncLock:
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> Optional[bool]:
        return False


class BaseStorage(AsyncContextManager):
    """
    >>> async with BaseStorage() as storage:
    >>>     async with storage.reader("/path/to/file", mode="b") as _reader:
    >>>         content = await _reader.read(1024)
    >>>
    >>>     async with storage.writer("/path/to/file", mode="b") as _writer:
    >>>         await _writer.write(b"12345")
    """
    EXIT_MAX_DELAY = 60
    _blocking_timeout = 600
    _write_timeout = 300
    _directory_create_mode = 0o755

    def __init__(self, **options):
        self.options = options
        self.blocking_timeout = self.options.get("blocking_timeout", self._blocking_timeout)
        self.write_timeout = self.options.get("write_timeout", self._write_timeout)
        self.directory_create_mode = self.options.get(
            "directory_create_mode", self._directory_create_mode
        )

    async def __aenter__(self) -> "BaseStorage":
        await self._connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> Optional[bool]:
        with anyio.fail_after(self.EXIT_MAX_DELAY, shield=True):
            await self._disconnect()

        return False

    def writer(self, path: str, mode: MODE = "b", append=False, **kwargs) -> "_AsyncWriter":
        return _AsyncWriter(self, path, mode, append=append, **kwargs)

    def reader(self, path: str, mode: MODE = "b", **kwargs) -> "_AsyncReader":
        return _AsyncReader(self, path, mode, **kwargs)

    async def delete(self, path: str):
        raise NotSupportedError(details="Not supported for this storage.")

    async def get_url(self, path: str) -> str:
        raise NotSupportedError(details="Not supported for this storage.")

    async def listdir(self, path: str) -> List[str]:
        raise NotSupportedError(details="Not supported for this storage.")

    async def exists(self, path: str) -> bool:
        raise NotSupportedError(details="Not supported for this storage.")

    async def size(self, path: str) -> int:
        raise NotSupportedError(details="Not supported for this storage.")

    async def get_mtime(self, path) -> float:
        raise NotSupportedError(details="Not supported for this storage.")

    # Protected methods, not to be used directly

    async def _connect(self):
        pass

    async def _disconnect(self):
        pass

    async def _open(self, path: str, mode="b", **kwargs) -> Any:
        raise NotSupportedError(details="Not supported for this storage.")

    async def _close(self, fd: Any) -> None:
        raise NotSupportedError(details="Not supported for this storage.")

    async def _write(self, fd: Any, stream: IO) -> None:
        raise NotSupportedError(details="Not supported for this storage.")

    async def _read(self, fd: Any, size: int = -1) -> AnyStr:
        raise NotSupportedError(details="Not supported for this storage.")

    async def _readline(self, fd: Any, size: int = -1) -> AnyStr:
        raise NotSupportedError(details="Not supported for this storage.")

    async def _finalize_read(self, fd: Any) -> None:
        pass

    async def _finalize_write(self, fd: Any) -> None:
        pass

    def get_access_lock(self, path: str, mode="r") -> AsyncContextManager:
        # Mode parameter allows possible usage of RWLock,
        # should you find a working cross-process implementation
        return DummyAsyncLock()


class _AsyncResourse(AsyncContextManager):
    def __init__(self, storage: BaseStorage, path: str, mode: str, **kwargs):
        self._storage = storage
        self._path = path
        self._mode = mode
        self._fd: Any = None
        self._kwargs = kwargs
        self._resource_lock = self._storage.get_access_lock(path, mode=mode)

        if self._mode not in ["t", "b"]:
            raise NotSupportedError(
                details="Supported modes for opening file are 't', 'b'."
            )

    async def __aenter__(self):
        _ = await self._resource_lock.__aenter__()
        self._fd = await self._storage._open(self._path, self._mode, **self._kwargs)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> Optional[bool]:
        await self.close_task()
        self._fd = None
        return False

    async def close_task(self):
        try:
            if self._fd:
                await self._storage._close(self._fd)
        finally:
            await self._resource_lock.__aexit__(*sys.exc_info())


class _AsyncWriter(_AsyncResourse):
    def __init__(self, storage: BaseStorage, path: str, mode: str, append=False, **kwargs):
        super().__init__(storage, path, mode, **kwargs)
        self._mode = ("a" if append else "w") + self._mode
        self._rollback = False

    async def write(self, content: IO) -> None:
        await self._storage._write(self._fd, content)

    async def close_task(self):
        try:
            if self._fd:
                await self._storage._finalize_write(self._fd)
        finally:
            await super().close_task()


class _AsyncReader(_AsyncResourse):
    def __init__(self, storage: BaseStorage, path: str, mode: str, **kwargs):
        super().__init__(storage, path, mode, **kwargs)
        self._mode = "r" + self._mode

    async def read(self, size: int = -1) -> AnyStr:
        return await self._storage._read(self._fd, size)

    async def readline(self, size: int = -1) -> AnyStr:
        return await self._storage._readline(self._fd, size)

    async def __aiter__(self) -> AsyncIterator[AnyStr]:
        while line := (await self.readline()):
            yield line

    async def close_task(self):
        try:
            if self._fd:
                await self._storage._finalize_read(self._fd)
        finally:
            await super().close_task()



import base64
import json
import os
import uuid
from datetime import datetime, timedelta

import pytest
from sqlalchemy.ext.asyncio import AsyncSession

from starlette_web.common.conf import settings
from starlette_web.contrib.auth.models import User, UserSession, UserInvite
from starlette_web.contrib.auth.utils import (
    decode_jwt,
    encode_jwt,
    TOKEN_TYPE_RESET_PASSWORD,
    TOKEN_TYPE_REFRESH,
    TOKEN_TYPE_ACCESS,
)
from starlette_web.tests.core.test_base import BaseTestAPIView
from starlette_web.tests.helpers import await_

INVALID_SIGN_IN_DATA = [
    [{"email": "fake-email"}, {"email": "Not a valid email address."}],
    [{"password": ""}, {"password": "Length must be between 2 and 32."}],
    [
        {},
        {
            "email": "Missing data for required field.",
            "password": "Missing data for required field.",
        },
    ],
]

INVALID_SIGN_UP_DATA = [
    [
        {},
        {
            "email": "Missing data for required field.",
            "password_1": "Missing data for required field.",
            "password_2": "Missing data for required field.",
            "invite_token": "Missing data for required field.",
        },
    ],
    [
        {"email": ("user_" * 30 + "@t.com")},
        {"email": "Longer than maximum length 128."},
    ],
    [
        {"email": "fake-email"},
        {"email": "Not a valid email address."},
    ],
    [
        {
            "email": "test@test.com",
            "invite_token": uuid.uuid4().hex,
            "password_1": "Head",
            "password_2": "Foo",
        },
        {"password_1": "Passwords must be equal", "password_2": "Passwords must be equal"},
    ],
    [
        {"invite_token": "token"},
        {"invite_token": "Length must be between 10 and 32."},
    ],
]
INVALID_INVITE_DATA = [
    [{"email": "fake-email"}, {"email": "Not a valid email address."}],
    [{}, {"email": "Missing data for required field."}],
]
INVALID_CHANGE_PASSWORD_DATA = [
    [{"password_1": "123456", "token": "t"}, {"password_2": "Missing data for required field."}],
    [{"password_1": "foo", "password_2": "foo"}, {"token": "Missing data for required field."}],
    [
        {"password_1": "header", "password_2": "footer", "token": "t"},
        {"password_1": "Passwords must be equal", "password_2": "Passwords must be equal"},
    ],
    [
        {"password_1": "header", "password_2": "footer", "token": ""},
        {"token": "Shorter than minimum length 1."},
    ],
]


def assert_tokens(response_data: dict, user: User, session_id: str = None):
    """Allows to check access- and refresh-tokens in the response body"""

    access_token = response_data.get("access_token")
    refresh_token = response_data.get("refresh_token")
    assert access_token, f"No access_token in response: {response_data}"
    assert refresh_token, f"No refresh_token in response: {response_data}"

    decoded_access_token = decode_jwt(access_token)
    access_exp_dt = datetime.fromisoformat(decoded_access_token.pop("exp_iso"))
    assert access_exp_dt > datetime.utcnow()
    assert decoded_access_token.get("user_id") == user.id, decoded_access_token
    assert decoded_access_token.get("token_type") == "access", decoded_access_token

    decoded_refresh_token = decode_jwt(refresh_token)
    refresh_exp_dt = datetime.fromisoformat(decoded_refresh_token.pop("exp_iso"))
    assert refresh_exp_dt > datetime.utcnow()
    assert decoded_refresh_token.get("user_id") == user.id, decoded_refresh_token
    assert decoded_refresh_token.get("token_type") == "refresh", decoded_refresh_token
    assert refresh_exp_dt > access_exp_dt

    if session_id:
        assert decoded_refresh_token.get("session_id") == session_id


class TestAuthMeAPIView(BaseTestAPIView):
    url = "/api/auth/me/"

    def test_get__ok(self, client, user):
        client.login(user)
        response = client.get(self.url)
        response_data = self.assert_ok_response(response)
        assert response_data == {
            "id": user.id,
            "email": user.email,
            "is_active": True,
            "is_superuser": user.is_superuser,
        }


class TestAuthSignInAPIView(BaseTestAPIView):
    url = "/api/auth/sign-in/"
    raw_password = "test-password"
    default_fail_status_code = 401

    @classmethod
    def setup_class(cls):
        cls.encoded_password = User.make_password(cls.raw_password)

    def setup_method(self):
        self.email = f"user_{uuid.uuid4().hex[:10]}@test.com"

    def _create_user(self, db_session, is_active=True):
        self.user = await_(
            User.async_create(
                db_session,
                db_commit=True,
                email=self.email,
                password=self.encoded_password,
                is_active=is_active,
            )
        )

    def test_sign_in__ok(self, client, dbs):
        self._create_user(dbs)
        response = client.post(self.url, json={"email": self.email, "password": self.raw_password})
        response_data = self.assert_ok_response(response)
        assert_tokens(response_data, self.user)

    def test_sign_in__check_user_session__ok(self, client, dbs):
        self._create_user(dbs)
        response = client.post(self.url, json={"email": self.email, "password": self.raw_password})
        response_data = self.assert_ok_response(response)

        refresh_token = response_data.get("refresh_token")
        decoded_refresh_token = decode_jwt(refresh_token)
        refresh_exp_dt = datetime.fromisoformat(decoded_refresh_token.pop("exp_iso"))

        user_session: UserSession = await_(UserSession.async_get(dbs, user_id=self.user.id))
        assert user_session.refresh_token == refresh_token
        assert user_session.is_active is True
        assert user_session.expired_at == refresh_exp_dt
        assert user_session.refreshed_at is not None
        assert decoded_refresh_token.get("session_id") == user_session.public_id

    def test_sign_in__create_new_user_session__ok(self, client, dbs):
        self._create_user(dbs)
        old_expired_at = datetime.now() + timedelta(seconds=1)
        old_user_session = await_(
            UserSession.async_create(
                dbs,
                is_active=True,
                user_id=self.user.id,
                public_id=str(uuid.uuid4()),
                refresh_token="refresh_token",
                expired_at=old_expired_at,
            )
        )
        response = client.post(self.url, json={"email": self.email, "password": self.raw_password})
        response_data = self.assert_ok_response(response)
        refresh_token = response_data.get("refresh_token")

        user_sessions: list[UserSession] = await_(
            UserSession.async_filter(dbs, user_id=self.user.id)
        ).all()
        assert len(user_sessions) == 2
        old_session, new_session = user_sessions

        assert old_session.id == old_user_session.id
        assert old_session.is_active is True
        assert old_session.expired_at == old_expired_at
        assert old_session.refreshed_at == old_user_session.refreshed_at

        assert new_session.refresh_token == refresh_token
        assert new_session.is_active is True
        assert old_session.refreshed_at is not None

    def test_sign_in__password_mismatch__fail(self, client, dbs):
        self._create_user(dbs)
        response = client.post(self.url, json={"email": self.email, "password": "fake-password"})
        response_data = self.assert_fail_response(response)
        assert response_data == {
            "error": "Authentication credentials are invalid.",
            "details": "Email or password is invalid.",
        }

    def test_sign_in__user_not_found__fail(self, client):
        response = client.post(self.url, json={"email": "fake@t.ru", "password": self.raw_password})
        response_data = self.assert_fail_response(response)
        assert response_data == {
            "error": "Authentication credentials are invalid.",
            "details": "Not found active user with provided email.",
        }

    @pytest.mark.parametrize("invalid_data, error_details", INVALID_SIGN_IN_DATA)
    def test_sign_in__invalid_request__fail(self, client, invalid_data: dict, error_details: dict):
        self.assert_bad_request(client.post(self.url, json=invalid_data), error_details)

    def test_sign_in__user_inactive__fail(self, client, dbs):
        self._create_user(dbs, is_active=False)
        response = client.post(self.url, json={"email": self.email, "password": self.raw_password})
        response_data = self.assert_fail_response(response)
        assert response_data == {
            "error": "Authentication credentials are invalid.",
            "details": "Not found active user with provided email.",
        }


class TestAuthSignUPAPIView(BaseTestAPIView):
    url = "/api/auth/sign-up/"
    default_fail_status_code = 400

    @staticmethod
    def _sign_up_data(user_invite: UserInvite):
        return {
            "email": user_invite.email,
            "invite_token": user_invite.token,
            "password_1": "test-password",
            "password_2": "test-password",
        }

    def test_sign_up__ok(self, client, user_invite, dbs):
        request_data = self._sign_up_data(user_invite)
        response = client.post(self.url, json=request_data)
        response_data = self.assert_ok_response(response, status_code=201)

        user = await_(User.async_get(dbs, email=request_data["email"]))
        assert user is not None, f"User wasn't created with {request_data=}"
        assert_tokens(response_data, user)

        await_(dbs.refresh(user_invite))
        assert user_invite.user_id == user.id
        assert user_invite.is_applied

    @pytest.mark.parametrize("invalid_data, error_details", INVALID_SIGN_UP_DATA)
    def test_sign_up__invalid_request__fail(self, client, invalid_data: dict, error_details: dict):
        self.assert_bad_request(client.post(self.url, json=invalid_data), error_details)

    def test_sign_up__user_already_exists__fail(self, client, user_invite, dbs):
        request_data = self._sign_up_data(user_invite)
        user_email = request_data["email"]

        await_(User.async_create(dbs, db_commit=True, email=user_email, password="pass"))
        response = client.post(self.url, json=request_data)
        response_data = self.assert_fail_response(response, status_code=409)
        assert response_data == {
            "error": "Request conflicts with current state of server.",
            "details": f"User with email '{user_email}' already exists",
        }

    @pytest.mark.parametrize(
        "token_update_data",
        [
            {"token": f"outdated-token-{uuid.uuid4().hex[:10]}"},
            {"expired_at": datetime.utcnow() - timedelta(hours=1)},
            {"is_applied": True},
        ],
    )
    def test_sign_up__token_problems__fail(self, client, user_invite, token_update_data, dbs):
        request_data = self._sign_up_data(user_invite)
        await_(user_invite.update(dbs, **token_update_data))
        await_(dbs.commit())
        response = client.post(self.url, json=request_data)
        response_data = self.assert_fail_response(response, status_code=422)
        assert response_data == {
            "error": "Could not process request due to logical errors in data.",
            "details": "Invitation link is expired or unavailable",
        }

    def test_sign_up__email_mismatch_with_token__fail(self, client, user_invite):
        request_data = self._sign_up_data(user_invite)
        request_data["email"] = f"another.email{uuid.uuid4().hex[:10]}@test.com"
        response = client.post(self.url, json=request_data)
        response_data = self.assert_fail_response(response)
        assert response_data["error"] == "Email does not match with your invitation."


class TestSignOutAPIView(BaseTestAPIView):
    url = "/api/auth/sign-out/"

    def test_sign_out__ok(self, client, user, dbs):
        user_session = client.login(user)
        response = client.delete(self.url)
        assert response.status_code == 200
        user_session = await_(UserSession.async_get(dbs, id=user_session.id))
        assert user_session.is_active is False

    def test_sign_out__user_session_not_found__ok(self, client, user):
        client.login(user)
        response = client.delete(self.url)
        assert response.status_code == 200


class TestUserInviteApiView(BaseTestAPIView):
    url = "/api/auth/invite-user/"
    default_fail_status_code = 400

    def setup_method(self):
        self.email = f"user_{uuid.uuid4().hex[:10]}@test.com"

    def test_invite__ok(self, client, user, mocked_auth_send, dbs):
        client.login(user)
        response = client.post(self.url, json={"email": self.email})
        response_data = self.assert_ok_response(response, status_code=201)

        user_invite: UserInvite = await_(UserInvite.async_get(dbs, email=self.email))
        assert user_invite is not None
        assert response_data == {
            "id": user_invite.id,
            "token": user_invite.token,
            "email": user_invite.email,
            "owner_id": user.id,
            "created_at": user_invite.created_at.isoformat(),
            "expired_at": user_invite.expired_at.isoformat(),
        }
        invite_data = {
            "token": user_invite.token,
            "email": user_invite.email,
        }
        invite_data = base64.urlsafe_b64encode(json.dumps(invite_data).encode()).decode()
        link = f"{settings.SITE_URL}/sign-up/?i={invite_data}"
        expected_body = (
            f"<p>Hello! :) You have been invited to {settings.SITE_URL}</p>"
            f"<p>Please follow the link </p>"
            f"<p><a href={link}>{link}</a></p>"
        )
        mocked_auth_send.assert_awaited_once_with(
            recipients_list=[self.email],
            subject=f"Welcome to {settings.SITE_URL}",
            html_content=expected_body,
        )

    @pytest.mark.parametrize("invalid_data, error_details", INVALID_INVITE_DATA)
    def test_invalid_request__fail(self, client, user, invalid_data: dict, error_details: dict):
        client.login(user)
        self.assert_bad_request(client.post(self.url, json=invalid_data), error_details)

    def test_invite__unauth__fail(self, client):
        client.logout()
        self.assert_unauth(client.post(self.url, json={"email": self.email}))

    def test_invite__user_already_exists__fail(self, client, user):
        client.login(user)
        response = client.post(self.url, json={"email": user.email})

        response_data = self.assert_fail_response(response)
        assert response_data == {
            "error": "Requested data is not valid.",
            "details": f"User with email=[{user.email}] already exists.",
        }

    # TODO: test
    def _test_invite__user_already_invited__update_invite__ok(
        self, client, user, mocked_auth_send, db_session
    ):
        old_token = UserInvite.generate_token()
        old_expired_at = datetime.utcnow()
        user_invite = await_(
            UserInvite.async_create(
                db_session,
                email=self.email,
                token=old_token,
                expired_at=old_expired_at,
                owner_id=user.id,
            )
        )

        client.login(user)
        client.post(self.url, json={"email": self.email})
        updated_user_invite: UserInvite = await_(UserInvite.async_get(db_session, email=self.email))

        assert updated_user_invite is not None
        assert updated_user_invite.id == user_invite.id
        assert updated_user_invite.token != user_invite.token
        assert updated_user_invite.expired_at != user_invite.expired_at
        assert mocked_auth_send.assert_awaited_once


class TestResetPasswordAPIView(BaseTestAPIView):
    url = "/api/auth/reset-password/"

    def setup_method(self):
        self.email = f"user_{uuid.uuid4().hex[:10]}@test.com"

    def test_reset_password__ok(self, client, user, mocked_auth_send, dbs):
        request_user = user
        await_(request_user.update(dbs, is_superuser=True))
        target_user = await_(
            User.async_create(dbs, db_commit=True, email=self.email, password="pass")
        )

        client.login(user)
        response = client.post(self.url, json={"email": target_user.email})
        response_data = self.assert_ok_response(response)
        token = response_data.get("token")

        assert response_data["user_id"] == target_user.id
        assert token is not None, response_data
        assert decode_jwt(response_data["token"])["user_id"] == target_user.id

        link = f"{settings.SITE_URL}/change-password/?t={token}"
        expected_body = (
            f"<p>You can reset your password for {settings.SITE_URL}</p>"
            f"<p>Please follow the link </p>"
            f"<p><a href={link}>{link}</a></p>"
        )
        mocked_auth_send.assert_awaited_once_with(
            recipients_list=[target_user.email],
            subject=f"Welcome back to {settings.SITE_URL}",
            html_content=expected_body,
        )

    def test_reset_password__unauth__fail(self, client):
        client.logout()
        self.assert_unauth(client.post(self.url, json={"email": self.email}))

    def test_reset_password__user_not_found__fail(self, client, user, mocked_auth_send, dbs):
        request_user = user
        await_(request_user.update(dbs, is_superuser=True))
        await_(dbs.commit())

        client.login(request_user)
        response = client.post(self.url, json={"email": "fake-email@test.com"})
        response_data = self.assert_fail_response(
            response, status_code=400,
        )
        assert response_data == {
            "error": "Requested data is not valid.",
            "details": "User with email=[fake-email@test.com] not found.",
        }

    def test_reset_password__user_is_not_superuser__fail(self, client, user):
        client.login(user)
        response = client.post(self.url, json={"email": user.email})
        response_data = self.assert_fail_response(
            response, status_code=403,
        )
        assert response_data == {
            "error": "You do not have permission to perform this action.",
            "details": "You don't have an admin privileges.",
        }


class TestChangePasswordAPIView(BaseTestAPIView):
    url = "/api/auth/change-password/"
    new_password = "new123456"

    def _assert_fail_response(self, client, token: str) -> dict:
        request_data = {
            "token": token,
            "password_1": self.new_password,
            "password_2": self.new_password,
        }
        client.logout()
        response = client.post(self.url, json=request_data)
        assert response.status_code == 401
        return response.json()

    def test_change_password__ok(self, client, user, user_session, dbs):
        token, _ = encode_jwt(
            {"user_id": user.id, "session_id": user_session.public_id},
            token_type=TOKEN_TYPE_RESET_PASSWORD,
        )
        request_data = {
            "token": token,
            "password_1": self.new_password,
            "password_2": self.new_password,
        }
        client.logout()
        response = client.post(self.url, json=request_data)
        response_data = self.assert_ok_response(response)
        assert_tokens(response_data, user)

        await_(dbs.refresh(user))
        assert user.verify_password(self.new_password)

    @pytest.mark.parametrize("invalid_data, error_details", INVALID_CHANGE_PASSWORD_DATA)
    def test_invalid_request__fail(self, client, invalid_data: dict, error_details: dict):
        client.logout()
        self.assert_bad_request(client.post(self.url, json=invalid_data), error_details)

    def test__token_expired__fail(self, client, user):
        token, _ = encode_jwt({"user_id": user.id}, expires_in=-10)
        response_data = self._assert_fail_response(client, token)
        self.assert_auth_invalid(
            response_data,
            "JWT signature has been expired for token",
            message="Authentication credentials have expired.",
        )

    def test__token_invalid_type__fail(self, client, user):
        token, _ = encode_jwt({"user_id": user.id}, token_type=TOKEN_TYPE_REFRESH)
        response_data = self._assert_fail_response(client, token)
        self.assert_auth_invalid(
            response_data, "Token type 'reset_password' expected, got 'refresh' instead."
        )

    def test_token_invalid__fail(self, client, user):
        response_data = self._assert_fail_response(client, "fake-jwt")
        self.assert_auth_invalid(response_data, "Token could not be decoded: Not enough segments")

    def test_user_inactive__fail(self, client, user, dbs):
        await_(user.update(dbs, is_active=False))
        await_(dbs.commit())
        token, _ = encode_jwt({"user_id": user.id}, token_type=TOKEN_TYPE_RESET_PASSWORD)
        response_data = self._assert_fail_response(client, token)
        self.assert_auth_invalid(response_data, f"Couldn't found active user with id={user.id}.")

    def test_user_does_not_exist__fail(self, client, user):
        user_id = 0
        token, _ = encode_jwt({"user_id": user_id}, token_type=TOKEN_TYPE_RESET_PASSWORD)
        response_data = self._assert_fail_response(client, token)
        self.assert_auth_invalid(response_data, f"Couldn't found active user with id={user_id}.")


class TestRefreshTokenAPIView(BaseTestAPIView):
    url = "/api/auth/refresh-token/"

    INVALID_REFRESH_TOKEN_DATA = [
        [{}, {"refresh_token": "Missing data for required field."}],
        [{"refresh_token": ""}, {"refresh_token": "Length must be between 10 and 512."}],
    ]

    @staticmethod
    def _prepare_token(
        db_session: AsyncSession, user: User, is_active=True, refresh=True
    ) -> UserSession:
        token_type = TOKEN_TYPE_REFRESH if refresh else TOKEN_TYPE_ACCESS
        session_id = str(uuid.uuid4())
        refresh_token, _ = encode_jwt(
            {"user_id": user.id, "session_id": session_id}, token_type=token_type
        )
        user_session = await_(
            UserSession.async_create(
                db_session,
                db_commit=True,
                user_id=user.id,
                is_active=is_active,
                public_id=session_id,
                refresh_token=refresh_token,
                expired_at=datetime.utcnow(),
            )
        )
        return user_session

    def test_refresh_token__ok(self, client, user, dbs):
        user_session = self._prepare_token(dbs, user)
        client.logout()
        response = client.post(self.url, json={"refresh_token": user_session.refresh_token})
        response_data = self.assert_ok_response(response)
        assert_tokens(response_data, user)

    def test_refresh_token__several_sessions_for_user__ok(self, client, user, dbs):
        user_session_1 = self._prepare_token(dbs, user)
        user_session_2 = self._prepare_token(dbs, user)
        client.logout()
        response = client.post(self.url, json={"refresh_token": user_session_2.refresh_token})
        response_data = self.assert_ok_response(response)
        assert_tokens(response_data, user, session_id=user_session_2.public_id)

        upd_user_session_1: UserSession = await_(UserSession.async_get(dbs, id=user_session_1.id))
        upd_user_session_2: UserSession = await_(UserSession.async_get(dbs, id=user_session_2.id))

        assert user_session_1.refreshed_at == upd_user_session_1.refreshed_at
        assert user_session_1.refresh_token == upd_user_session_1.refresh_token

        if os.name == "nt":
            # Windows has less time ticks, than Linux,
            # so sometimes these 2 times are equal up to milliseconds
            assert user_session_1.refreshed_at <= upd_user_session_2.refreshed_at
        else:
            assert user_session_1.refreshed_at < upd_user_session_2.refreshed_at

    def test_refresh_token__user_inactive__fail(self, client, user, dbs):
        user_session = self._prepare_token(dbs, user)
        await_(user.update(dbs, is_active=False))
        await_(dbs.commit())
        response = client.post(self.url, json={"refresh_token": user_session.refresh_token})
        self.assert_auth_invalid(response, f"Couldn't found active user with id={user.id}.")

    def test_refresh_token__session_inactive__fail(self, client, user, dbs):
        session = self._prepare_token(dbs, user, is_active=False)
        response = client.post(self.url, json={"refresh_token": session.refresh_token})
        expected_msg = (
            f"Couldn't found active session: user_id={user.id} | session_id='{session.public_id}'."
        )
        self.assert_auth_invalid(response, expected_msg)

    @pytest.mark.parametrize("token_type", [TOKEN_TYPE_ACCESS, TOKEN_TYPE_RESET_PASSWORD])
    def test_refresh_token__token_not_refresh__fail(self, client, user, token_type):
        refresh_token, _ = encode_jwt({"user_id": user.id}, token_type=token_type)
        response = client.post(self.url, json={"refresh_token": refresh_token})
        response_data = self.assert_fail_response(response, status_code=401)
        assert response_data == {
            "error": "Authentication credentials are invalid.",
            "details": f"Token type 'refresh' expected, got '{token_type}' instead.",
        }

    def test_refresh_token__token_mismatch__fail(self, client, user, dbs):
        user_session = self._prepare_token(dbs, user, is_active=True)
        refresh_token = user_session.refresh_token
        await_(user_session.update(dbs, refresh_token="fake-token"))
        await_(dbs.commit())
        response = client.post(self.url, json={"refresh_token": refresh_token})
        self.assert_auth_invalid(
            response, "Refresh token does not match with user session.",
        )

    def test_refresh_token__fake_jwt__fail(self, client, user):
        response = client.post(self.url, json={"refresh_token": "fake-jwt-token"})
        self.assert_auth_invalid(response, "Token could not be decoded: Not enough segments")

    @pytest.mark.parametrize("invalid_data, error_details", INVALID_REFRESH_TOKEN_DATA)
    def test_invalid_request__fail(self, client, invalid_data: dict, error_details: dict):
        client.logout()
        self.assert_bad_request(client.post(self.url, json=invalid_data), error_details)



import json

from apispec.ext.marshmallow import OpenAPIConverter
from apispec.ext.marshmallow.field_converter import DEFAULT_FIELD_MAPPING
import marshmallow

from starlette_web.common.conf import settings
from starlette_web.common.utils.json import StarletteJSONEncoder
from starlette_web.contrib.camel_case.utils import camelize, camelize_key


# TODO: maybe allow user to override class in settings (?)
class StarletteWebMarshmallowOpenAPIConverter(OpenAPIConverter):
    # Decimal has no specific representation in OpenAPI standard
    # Support decimal representation in a same way, as drf-yasg does for Django
    # https://github.com/axnsan12/drf-yasg/blob/b99306f71c6a5779b62189df7d9c1f5ea1c794ef/src/drf_yasg/openapi.py#L48  # noqa: E501
    field_mapping = {
        **DEFAULT_FIELD_MAPPING,
        marshmallow.fields.Decimal: ("string", "decimal"),
    }

    def field2choices(self, field, **kwargs):
        res = super().field2choices(field, **kwargs)
        return json.loads(StarletteJSONEncoder().encode(res))

    def fields2jsonschema(self, fields, *, partial=None):
        res = super().fields2jsonschema(fields, partial=partial)

        if settings.APISPEC["CONVERT_TO_CAMEL_CASE"]:
            if "required" in res and type(res["required"]) is list:
                res["required"] = [camelize_key(d) for d in res["required"]]
            if res.get("properties"):
                res["properties"] = camelize(res["properties"])
            return camelize(res)

        return res



# flake8: noqa E402
import os

os.environ.setdefault("STARLETTE_SETTINGS_MODULE", "starlette_web.tests.settings")

import asyncio
import sys
import uuid
from datetime import datetime, timedelta
from typing import Tuple
from unittest.mock import patch, AsyncMock

import pytest
import sqlalchemy
from sqlalchemy.engine import URL
from sqlalchemy.util import concurrency
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.exc import ProgrammingError, OperationalError

from starlette_web.common import database
from starlette_web.common.conf import settings
from starlette_web.contrib.auth.models import UserInvite
from starlette_web.tests.helpers import (
    WebTestClient,
    get_user_data,
    create_user,
    mock_target_class,
    create_user_session,
    make_db_session,
    await_,
)
from starlette_web.tests.mocks import MockProcess


if (*sys.version_info,) >= (3, 8) and sys.platform.startswith("win"):
    asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())


@pytest.fixture(autouse=True, scope="session")
def test_settings():
    settings.APP_DEBUG = True
    settings.MAX_UPLOAD_ATTEMPT = 1
    settings.RETRY_UPLOAD_TIMEOUT = 0


@pytest.fixture(autouse=True, scope="session")
def config():
    from starlette_web.contrib.constance import config as base_config

    settings.CONSTANCE_CONFIG = {
        "TEST_CONSTANT_1": (1, "Test constant 1", int),
        "TEST_CONSTANT_2": (2, "Test constant 2", int),
        "TEST_CONSTANT_UUID": (
            uuid.UUID("094eb5ff-01de-4985-afeb-22ebb9e76abf"),
            "Test constant uuid",
            uuid.UUID,
        ),
        "TEST_CONSTANT_DATETIME": (datetime(2000, 1, 1), "Test constant datetime", datetime),
    }
    settings.CONSTANCE_BACKEND = "starlette_web.contrib.constance.backends.database.DatabaseBackend"
    settings.CONSTANCE_DATABASE_CACHE_BACKEND = "locmem"

    base_config._setup()
    yield base_config


@pytest.fixture(autouse=True, scope="session")
def client() -> WebTestClient:
    from starlette_web.common.app import get_app

    with WebTestClient(get_app(run_checks_on_startup=False)) as client:
        with make_db_session(asyncio.get_event_loop_policy().get_event_loop()) as db_session:
            client.db_session = db_session
            yield client


@pytest.fixture
def dbs(loop) -> AsyncSession:
    with make_db_session(loop) as db_session:
        yield db_session


def db_prep():
    print("Dropping the old test db…")
    postgres_db_dsn = URL.create(
        drivername="postgresql",
        username=settings.DATABASE["username"],
        password=settings.DATABASE["password"],
        host=settings.DATABASE["host"],
        port=settings.DATABASE["port"],
        database="postgres",
    )
    engine = sqlalchemy.create_engine(postgres_db_dsn)
    conn = engine.connect()

    def exec_sql(query: str):
        conn.execute(sqlalchemy.text(query))

    try:
        conn = conn.execution_options(autocommit=False)
        exec_sql("ROLLBACK")
        exec_sql(f"DROP DATABASE {settings.DB_NAME}")
    except ProgrammingError:
        print("Could not drop the database, probably does not exist.")
        exec_sql("ROLLBACK")
    except OperationalError:
        print("Could not drop database because it’s being accessed by other users")
        exec_sql("ROLLBACK")

    print(f"Test db dropped! about to create {settings.DB_NAME}")
    exec_sql(f"CREATE DATABASE {settings.DB_NAME}")
    username, password = settings.DATABASE["username"], settings.DATABASE["password"]

    try:
        exec_sql(f"CREATE USER {username} WITH ENCRYPTED PASSWORD '{password}'")
    except Exception as e:
        print(f"User already exists. ({e})")
        exec_sql(f"GRANT ALL PRIVILEGES ON DATABASE {settings.DB_NAME} TO {username}")

    conn.close()


@pytest.fixture(autouse=True, scope="session")
def db_migration():
    def create_tables():
        db_prep()
        print("Creating tables...")
        engine = sqlalchemy.create_engine(settings.DATABASE_DSN)
        database.ModelBase.metadata.create_all(engine)

    await_(concurrency.greenlet_spawn(create_tables))
    print("DB and tables were successful created.")


@pytest.fixture
def mocked_process(monkeypatch) -> MockProcess:
    yield from mock_target_class(MockProcess, monkeypatch)


@pytest.fixture
def mocked_auth_send() -> AsyncMock:
    mocked_send_email = AsyncMock()
    patcher = patch("starlette_web.contrib.auth.views.send_email", new=mocked_send_email)
    patcher.start()
    yield mocked_send_email
    del mocked_send_email
    patcher.stop()


@pytest.fixture()
def user_data() -> Tuple[str, str]:
    return get_user_data()


@pytest.fixture
def loop():
    return asyncio.get_event_loop_policy().get_event_loop()


@pytest.fixture
def user(dbs):
    return create_user(dbs)


@pytest.fixture
def user_session(user, loop, dbs):
    return create_user_session(dbs, user)


@pytest.fixture
def user_invite(user, loop, dbs) -> UserInvite:
    return loop.run_until_complete(
        UserInvite.async_create(
            dbs,
            db_commit=True,
            email=f"user_{uuid.uuid4().hex[:10]}@test.com",
            token=f"{uuid.uuid4().hex}",
            expired_at=datetime.utcnow() + timedelta(days=1),
            owner_id=user.id,
        )
    )



import typing
from marshmallow import Schema, ValidationError
from webargs import fields, validate


__all__ = [
    "SignInSchema",
    "SignUpSchema",
    "JWTResponseSchema",
    "UserResponseSchema",
    "RefreshTokenSchema",
    "ChangePasswordSchema",
    "UserInviteRequestSchema",
    "UserInviteResponseSchema",
    "ResetPasswordRequestSchema",
    "ResetPasswordResponseSchema",
]


class TwoPasswordsMixin:
    @staticmethod
    def is_valid(data: typing.Mapping) -> typing.Mapping:
        if data["password_1"] != data["password_2"]:
            msg = "Passwords must be equal"
            raise ValidationError(msg, data={"password_1": msg, "password_2": msg})

        return data


class SignInSchema(Schema):
    email = fields.Email(required=True)
    password = fields.Str(required=True, validate=validate.Length(min=2, max=32))


class SignUpSchema(TwoPasswordsMixin, Schema):
    email = fields.Email(required=True, validate=validate.Length(max=128))
    password_1 = fields.Str(required=True, validate=validate.Length(min=2, max=32))
    password_2 = fields.Str(required=True, validate=validate.Length(min=2, max=32))
    invite_token = fields.Str(required=True, validate=validate.Length(min=10, max=32))


class RefreshTokenSchema(Schema):
    refresh_token = fields.Str(required=True, validate=validate.Length(min=10, max=512))


class JWTResponseSchema(Schema):
    access_token = fields.Str(required=True)
    refresh_token = fields.Str(required=True)

    class Meta:
        fields = ("access_token", "refresh_token")


class UserInviteRequestSchema(Schema):
    email = fields.Email(required=True)


class UserInviteResponseSchema(Schema):
    id = fields.Int()
    email = fields.Email(required=True)
    token = fields.Str(required=True)
    expired_at = fields.DateTime(required=True)
    created_at = fields.DateTime(required=True)
    owner_id = fields.Int(required=True)


class ResetPasswordRequestSchema(Schema):
    email = fields.Email(required=True)


class ResetPasswordResponseSchema(Schema):
    user_id = fields.Int()
    email = fields.Email(required=True)
    token = fields.Str(required=True)


class ChangePasswordSchema(TwoPasswordsMixin, Schema):
    token = fields.Str(required=True, validate=validate.Length(min=1))
    password_1 = fields.Str(required=True, validate=validate.Length(min=2, max=32))
    password_2 = fields.Str(required=True, validate=validate.Length(min=2, max=32))


class UserResponseSchema(Schema):
    id = fields.Int(required=True)
    email = fields.Email(required=True)
    is_active = fields.Bool(required=True)
    is_superuser = fields.Bool(required=True)



from starlette_web.common.management.alembic_mixin import AlembicMixin
from starlette_web.common.management.base import BaseCommand, CommandParser


class Command(BaseCommand, AlembicMixin):
    help = "Migrate to selected alembic revision"

    def add_arguments(self, parser: CommandParser):
        parser.add_argument(
            "name",
            type=str,
            help="Alembic revision slug to migrate to",
            nargs="?",
        )

    async def handle(self, **options):
        revision_name = options.get("name")

        if revision_name is None:
            stdout, stderr = await self.run_alembic_main(["upgrade", "head"])

        elif revision_name.startswith("+") and revision_name[1:].isdigit():
            stdout, stderr = await self.run_alembic_main(["upgrade", revision_name])

        elif revision_name.startswith("-") and revision_name[1:].isdigit():
            stdout, stderr = await self.run_alembic_main(["downgrade", revision_name])

        else:
            stdout, stderr = await self.run_alembic_main(["upgrade", revision_name])
            print(stdout.strip())
            print(stderr.strip())

            stdout, stderr = await self.run_alembic_main(["downgrade", revision_name])

        print(stdout.strip())
        print(stderr.strip())



import enum
from typing import Type, TypeVar

import sqlalchemy as sa
from sqlalchemy import Column
from sqlalchemy.sql import type_api as sa_type_api

from starlette_web.common.database.types import ChoiceType


EnumClass = TypeVar("EnumClass", bound=enum.Enum)


# TODO: probably not supported in starlette_admin,
#  one has to manually set field converter to sqlalchemy implementation
#  in starlette_admin.contrib.sqla.helpers.converters
class EnumTypeColumn(Column):
    """Just wrapper for ChoiceType db column

    >>> import enum
    >>> from sqlalchemy import String
    >>> from starlette_web.common.database import ModelBase

    >>> class UserType(enum.Enum):
    >>>    admin = 'admin'
    >>>    regular = 'regular'

    >>> class User(ModelBase):
    >>>     ...
    >>>     type = EnumTypeColumn(UserType, impl=String(16), default=UserType.admin)

    >>> user = User(type='admin')
    >>> user.type
    [0] 'admin'

    """

    impl = sa.String(16)

    def __new__(
        cls, enum_class: Type[EnumClass], impl: sa_type_api.TypeEngine = None, *args, **kwargs
    ):
        if "default" in kwargs:
            kwargs["default"] = getattr(kwargs["default"], "value") or kwargs["default"]

        impl = impl or cls.impl
        return Column(ChoiceType(enum_class, impl=impl), *args, **kwargs)



from apispec.ext.marshmallow import MarshmallowPlugin

from starlette_web.contrib.apispec.marshmallow.converters import (
    StarletteWebMarshmallowOpenAPIConverter,
)


# TODO: maybe allow user to override class in settings (?)
class StarletteWebMarshmallowPlugin(MarshmallowPlugin):
    Converter = StarletteWebMarshmallowOpenAPIConverter



from typing import Any, Callable, ByteString

from starlette_web.common.caches import caches
from starlette_web.common.conf import settings
from starlette_web.common.http.exceptions import ImproperlyConfigured
from starlette_web.contrib.constance.backends.base import BaseConstanceBackend


class ConstanceCacheMixin(BaseConstanceBackend):
    _cache_key = "default"
    _caching_timeout = None
    _preprocess_response: Callable[[ByteString], Any]

    def __init__(self):
        super().__init__()
        if self._cache_key not in settings.CACHES:
            raise ImproperlyConfigured(
                details="ConstanceCacheBackend must be configured "
                        "with existing cache_key in settings.CACHES"
            )
        self._cache = caches[self._cache_key]

    async def get(self, key: str) -> Any:
        value = self._preprocess_response(
            await self._cache.async_get(self._cache_keygen(key))
        )
        if value == self.empty:
            value = await super().get(key)
            if value != self.empty:
                await self._cache.async_set(
                    self._cache_keygen(key),
                    self.serializer.serialize(value),
                    timeout=self._caching_timeout,
                )
        return value

    async def set(self, key: str, value: Any) -> None:
        await self._cache.async_set(
            self._cache_keygen(key),
            self.serializer.serialize(value),
            timeout=self._caching_timeout,
        )
        await super().set(key, value)

    @staticmethod
    def _cache_keygen(key: str) -> str:
        return ":constance:" + key



# Copy from https://github.com/django/django/blob/main/django/utils/crypto.py

import secrets
import string
from typing import Union


RANDOM_STRING_CHARS = string.ascii_letters + string.digits


def _force_bytes(value: Union[str, bytes], encoding="utf-8", errors="strict"):
    if isinstance(value, bytes):
        if encoding == "utf-8":
            return value
        else:
            return value.decode("utf-8", errors).encode(encoding, errors)
    if isinstance(value, memoryview):
        return bytes(value)
    return str(value).encode(encoding, errors)


def get_random_string(length, allowed_chars=RANDOM_STRING_CHARS):
    """
    Return a securely generated random string.
    The bit length of the returned value can be calculated with the formula:
        log_2(len(allowed_chars)^length)
    For example, with default `allowed_chars` (26+26+10), this gives:
      * length: 12, bit length =~ 71 bits
      * length: 22, bit length =~ 131 bits
    """
    return "".join([secrets.choice(allowed_chars) for _ in range(length)])


def constant_time_compare(val1, val2):
    """Return True if the two strings are equal, False otherwise."""
    return secrets.compare_digest(_force_bytes(val1), _force_bytes(val2))



import math
import sys
from typing import Optional

import anyio
from anyio._core._tasks import TaskGroup

from starlette_web.common.caches.base import CacheLockError
from starlette_web.common.http.exceptions import NotSupportedError


class BaseLock:
    EXIT_MAX_DELAY = 60.0

    def __init__(
        self,
        name: str,
        timeout: Optional[float] = None,
        blocking_timeout: Optional[float] = None,
        **kwargs,
    ) -> None:
        self._name = name
        if timeout is None:
            timeout = math.inf
        self._timeout = timeout
        if self._timeout is not None and self._timeout < 0:
            raise RuntimeError("timeout cannot be negative")

        self._blocking_timeout = blocking_timeout
        if self._blocking_timeout is not None and self._blocking_timeout < 0:
            raise RuntimeError("blocking_timeout cannot be negative")

        self._task_group_wrapper: Optional[TaskGroup] = None
        self._task_group: Optional[TaskGroup] = None
        self._acquire_event: Optional[anyio.Event] = None
        self._is_acquired = False

    async def __aenter__(self):
        self._task_group_wrapper = anyio.create_task_group()
        self._task_group = await self._task_group_wrapper.__aenter__()
        self._acquire_event = anyio.Event()
        if self._blocking_timeout is not None:
            self._task_group.cancel_scope.deadline = anyio.current_time() + self._blocking_timeout
        self._task_group.start_soon(self._acquire)

        try:
            await self._acquire_event.wait()
            self._is_acquired = self._acquire_event.is_set()
        except anyio.get_cancelled_exc_class() as exc:
            await self._task_group_wrapper.__aexit__(*sys.exc_info())
            self._is_acquired = False
            raise CacheLockError(details=str(exc)) from exc

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            with anyio.move_on_after(self.EXIT_MAX_DELAY, shield=True):
                await self._release()
        finally:
            retval = await self._task_group_wrapper.__aexit__(exc_type, exc_val, exc_tb)

        try:
            if exc_type is not None and exc_type not in [
                CacheLockError,
                anyio.get_cancelled_exc_class(),
            ]:
                # The lock itself is supposed to always raise CacheLockError on any inner error.
                # Furthermore, lock may be cancelled from outside with CancelledError.
                # Any other error is propagated.
                retval = False

            elif self._task_group.cancel_scope.cancel_called:
                raise CacheLockError(
                    message=f"Could not acquire FileLock within {self._timeout} seconds.",
                    details=str(sys.exc_info()[1]),
                ) from exc_val
        finally:
            self._acquire_event = None
            self._task_group = None
            self._task_group_wrapper = None

        return retval

    async def _acquire(self):
        raise NotSupportedError(details=f"{self.__class__.__name__} does not support _acquire")

    async def _release(self):
        raise NotSupportedError(details=f"{self.__class__.__name__} does not support _release")



import anyio

from starlette_web.common.conf import settings
from starlette_web.common.caches import caches
from starlette_web.common.channels.base import Channel
from starlette_web.common.management.base import BaseCommand
from starlette_web.contrib.redis.channel_layers import RedisPubSubChannelLayer


class Command(BaseCommand):
    help = "Command to cross-process test channels subscriber"

    def add_arguments(self, parser):
        parser.add_argument("--group", type=str, required=True)
        parser.add_argument("--subscriber", type=str, required=True)

    async def handle(self, **options):
        group = options["group"]
        subscriber_name = options["subscriber"]
        redis_options = settings.CHANNEL_LAYERS["redispubsub"]["OPTIONS"]
        channel_ctx = Channel(RedisPubSubChannelLayer(**redis_options))

        messages = []

        with anyio.move_on_after(5):
            async with channel_ctx as channel:
                async with channel.subscribe(group) as subscriber:
                    async for event in subscriber:
                        messages.append(event.message)
                        if event.message == "DONE":
                            break

        await caches["default"].async_set(f"{group}-{subscriber_name}-done", messages, 10)



# Copy from https://github.com/django/django/blob/main/django/core/cache/__init__.py

from typing import Dict, Type

from starlette_web.common.conf import settings
from starlette_web.common.caches.base import BaseCache, CacheError
from starlette_web.common.utils import import_string


def _create_cache(alias: str) -> BaseCache:
    try:
        cache_class: Type[BaseCache] = import_string(settings.CACHES[alias]["BACKEND"])
        return cache_class(settings.CACHES[alias]["OPTIONS"])
    except (ImportError, KeyError) as exc:
        raise CacheError from exc


class CacheHandler:
    def __init__(self):
        self._caches: Dict[str, BaseCache] = dict()

    def __getitem__(self, alias):
        try:
            return self._caches[alias]
        except KeyError:
            settings_caches = getattr(settings, "CACHES", {})
            if alias not in settings_caches:
                raise CacheError(f"Cache {alias} not in settings.CACHES")

            self._caches[alias] = _create_cache(alias)
            return self._caches[alias]


caches = CacheHandler()



from unittest.mock import patch

from starlette_web.tests.core.test_base import BaseTestAPIView


class TestHealthCheckAPIView(BaseTestAPIView):
    url = "/health_check/"

    def test_health__ok(self, client):
        response = client.get(self.url)
        response_data = self.assert_ok_response(response)
        assert response_data == {"services": {"postgres": "ok"}, "errors": []}

    @patch("starlette_web.common.database.ModelMixin.async_filter")
    def test_health__fail(self, mock_filter, client):
        mock_filter.side_effect = RuntimeError("Oops")
        response = client.get(self.url)

        response_data = response.json()
        assert response.status_code == 503
        assert response_data == {
            "services": {"postgres": "down"},
            "errors": ["Couldn't connect to DB: RuntimeError 'Oops'"],
        }



from starlette.requests import Request
from starlette.responses import Response
from starlette_admin.auth import AdminUser, AuthProvider
from starlette_admin.exceptions import LoginFailed

from starlette_web.common.http.exceptions import AuthenticationFailedError, PermissionDeniedError
from starlette_web.contrib.auth.backend import SessionJWTAuthenticationBackend
from starlette_web.contrib.auth.views import JWTSessionMixin
from starlette_web.contrib.auth.models import User, UserSession
from starlette_web.contrib.auth.permissions import IsSuperuserPermission
from starlette_web.contrib.auth.utils import TOKEN_TYPE_REFRESH, decode_jwt


class AdminAuthProvider(AuthProvider):
    async def login(
        self,
        username: str,
        password: str,
        remember_me: bool,
        request: Request,
        response: Response,
    ) -> Response:
        user = await User.async_get(
            db_session=request.state.session,
            email=username,
            is_active__is=True,
        )
        if not user:
            raise LoginFailed("Invalid username or password")

        if not user.verify_password(password):
            raise LoginFailed("Invalid username or password")

        if not user.is_superuser:
            raise LoginFailed(PermissionDeniedError.message)

        token_maker = JWTSessionMixin()
        token_maker.db_session = request.state.session
        token_collection = await token_maker._create_session(user)

        await UserSession.async_update(
            db_session=request.state.session,
            filter_kwargs=dict(refresh_token=token_collection.refresh_token),
            update_data=dict(is_persistent=remember_me),
            db_commit=True,
        )

        request.scope["session"] = {
            "token": token_collection.refresh_token,
            **decode_jwt(token_collection.refresh_token),
        }
        return response

    async def is_authenticated(self, request) -> bool:
        try:
            await SessionJWTAuthenticationBackend(request, request.scope,).authenticate(
                token_type=TOKEN_TYPE_REFRESH,
            )
            return await IsSuperuserPermission().has_permission(request, request.scope)
        except (AuthenticationFailedError, PermissionDeniedError):
            return False

    def get_admin_user(self, request: Request) -> AdminUser:
        if "user" in request.scope:
            username = request.scope["user"].email
        else:
            username = "Anonymous"
        return AdminUser(username=username, photo_url=None)

    async def logout(self, request: Request, response: Response) -> Response:
        request.scope["session"] = {}
        return response



import time

import anyio

from starlette_web.common.caches import caches
from starlette_web.tests.core.helpers.base_cache_tester import BaseCacheTester
from starlette_web.tests.helpers import await_


class TestRedisCache(BaseCacheTester):
    def test_redis_cache_base_ops(self):
        self._run_base_cache_test(caches["default"])

    def test_redis_cache_many_ops(self):
        self._run_cache_many_ops_test(caches["default"])

    def test_redis_lock(self):
        self._run_cache_lock_test(caches["default"])

    def test_redis_lock_race_condition(self):
        self._run_cache_mutual_lock_test(caches["default"])

    def test_redis_lock_correct_task_blocking(self):
        self._run_locks_timeouts_test(caches["default"])

    def test_redis_lock_cancellation(self):
        async def task_lock_cancel():
            with anyio.move_on_after(0.1):
                async with caches["default"].lock(
                    "test_lock_cancel",
                    timeout=100,
                    blocking_timeout=2,
                ):
                    await anyio.sleep(100)

        start_time = time.time()
        await_(task_lock_cancel())
        end_time = time.time()
        run_time = end_time - start_time
        assert (run_time - 0.1) < 0.05

        key = await_(caches["default"].async_get("test_lock_cancel"))
        assert key is None



from typing import List, Dict, Any, Type, ByteString

from starlette_web.common.http.exceptions import NotSupportedError, UnexpectedError
from starlette_web.common.utils.serializers import (
    BytesSerializer,
    PickleSerializer,
    DeserializeError,
)


class BaseConstanceBackend:
    serializer_class: Type[BytesSerializer] = PickleSerializer

    class EmptyValue:
        def __eq__(self, other):
            return self.__class__ == other.__class__

    empty = EmptyValue()

    def __init__(self):
        self.serializer: BytesSerializer = self.serializer_class()

    def _preprocess_response(self, response: ByteString) -> Any:
        if response is None:
            return self.empty

        if isinstance(response, (memoryview, bytearray)):
            response = bytes(response)

        try:
            return self.serializer.deserialize(response)
        except DeserializeError as exc:
            raise UnexpectedError(details=str(exc)) from exc

    async def mget(self, keys: List[str]) -> Dict[str, Any]:
        raise NotSupportedError(details="Constance backend not initialized.")

    async def get(self, key: str) -> Any:
        raise NotSupportedError(details="Constance backend not initialized.")

    async def set(self, key: str, value: Any) -> None:
        raise NotSupportedError(details="Constance backend not initialized.")



import secrets
from datetime import datetime

from sqlalchemy import Column, Integer, String, DateTime, Boolean, ForeignKey
from sqlalchemy.ext.asyncio import AsyncSession

from starlette_web.common.database import ModelMixin, ModelBase
from starlette_web.common.authorization.base_user import BaseUserMixin
from starlette_web.contrib.auth.hashers import make_password, verify_password


class User(ModelBase, BaseUserMixin, ModelMixin):
    __tablename__ = "auth_users"

    id = Column(Integer, primary_key=True)
    email = Column(String(length=128), index=True, nullable=False, unique=True)
    password = Column(String(length=256), nullable=False)
    is_active = Column(Boolean, default=True)
    is_superuser = Column(Boolean, default=False)

    class Meta:
        order_by = ("id",)

    def __repr__(self):
        return f"<User #{self.id} {self.email}>"

    def __str__(self):
        return self.email

    @classmethod
    def make_password(cls, raw_password: str) -> str:
        return make_password(raw_password)

    def verify_password(self, raw_password: str) -> bool:
        return verify_password(raw_password, str(self.password))

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return self.email

    @classmethod
    async def get_active(cls, db_session: AsyncSession, user_id: int) -> "User":
        return await cls.async_get(db_session, id=user_id, is_active=True)


class UserInvite(ModelBase, ModelMixin):
    __tablename__ = "auth_invites"
    TOKEN_MAX_LENGTH = 32

    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey("auth_users.id"), unique=True)
    email = Column(String(length=128), unique=True)
    token = Column(String(length=32), unique=True, nullable=False, index=True)
    is_applied = Column(Boolean, default=False, nullable=False)
    expired_at = Column(DateTime, nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    owner_id = Column(Integer, ForeignKey("auth_users.id"), nullable=False)

    class Meta:
        order_by = ("id",)

    def __repr__(self):
        return f"<UserInvite #{self.id} {self.token}>"

    @classmethod
    def generate_token(cls):
        return secrets.token_urlsafe()[: cls.TOKEN_MAX_LENGTH]


class UserSession(ModelBase, ModelMixin):
    __tablename__ = "auth_sessions"

    id = Column(Integer, primary_key=True)
    public_id = Column(String(length=36), index=True, nullable=False, unique=True)
    user_id = Column(Integer, ForeignKey("auth_users.id"))
    refresh_token = Column(String(length=512))
    is_active = Column(Boolean, default=True, nullable=False)
    is_persistent = Column(Boolean, default=True, nullable=False)
    expired_at = Column(DateTime, nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    refreshed_at = Column(DateTime, default=datetime.utcnow, nullable=False)

    class Meta:
        order_by = ("id",)

    def __repr__(self):
        return f"<UserSession #{self.id} {self.user_id}>"



from typing import Optional, List, TypedDict, Dict, Type

from starlette_web.common.conf import settings
from starlette_web.common.email.base_sender import BaseEmailSender
from starlette_web.common.http.exceptions import ImproperlyConfigured
from starlette_web.common.utils.importing import import_string


class EmailSenderSettings(TypedDict):
    BACKEND: str
    OPTIONS: Dict


class EmailManager:
    sender: Optional[BaseEmailSender] = None

    def _switch_to_email_sender_class(self, sender_settings: EmailSenderSettings):
        try:
            backend: Type[BaseEmailSender] = import_string(sender_settings["BACKEND"])
            self.sender = backend(**sender_settings["OPTIONS"])
        except (ImportError, SystemError, TypeError):
            raise ImproperlyConfigured(
                "settings.EMAIL_SENDER is not a valid import path to a callable"
            )

        if not isinstance(self.sender, BaseEmailSender):
            raise ImproperlyConfigured(
                "Class, specified with settings.EMAIL_SENDER, "
                "is not a subclass of BaseEmailSender"
            )

    def _get_email_sender(self) -> BaseEmailSender:
        if self.sender:
            return self.sender

        self._switch_to_email_sender_class(settings.EMAIL_SENDER)
        return self.sender

    async def send_email(
        self,
        subject: str,
        html_content: str,
        recipients_list: List[str],
        from_email: Optional[str] = None,
    ):
        async with self._get_email_sender() as sender:
            await sender.send_email(subject, html_content, recipients_list, from_email=from_email)


_email_manager = EmailManager()
send_email = _email_manager.send_email



import time


def test_websocket_chat(client):
    response_pool_1 = []
    response_pool_2 = []

    with client.websocket_connect("/ws/test_websocket_chat") as websocket:
        with client.websocket_connect("/ws/test_websocket_chat") as websocket_2:

            def receive():
                nonlocal response_pool_1, response_pool_2
                response_pool_1.append(websocket.receive_json())
                response_pool_2.append(websocket_2.receive_json())

            websocket.send_json({"request_type": "connect"})

            # These messages will be sent to first websocket, which is already instantiated,
            # but second websocket will not receive them.
            websocket.send_json({"request_type": "publish", "message": "Unreceived message 1"})
            websocket.receive_json()
            websocket.send_json({"request_type": "publish", "message": "Unreceived message 2"})
            websocket.receive_json()
            websocket_2.send_json({"request_type": "connect"})

            # If send message immediately after
            # websocket_2 initialized, it may skip the message
            time.sleep(0.1)

            websocket.send_json({"request_type": "publish", "message": "Hello there!"})
            receive()

            websocket_2.send_json({"request_type": "publish", "message": "General Kenobi!"})
            receive()

            websocket_2.send_json(
                {"request_type": "publish", "message": "I will deal with this Jedi slime myself."}
            )
            receive()

            websocket.send_json({"request_type": "publish", "message": "Your move."})
            receive()

            websocket_2.close(1001)

        websocket.close(1001)

    all_responses = [
        "Hello there!",
        "General Kenobi!",
        "I will deal with this Jedi slime myself.",
        "Your move.",
    ]
    assert response_pool_1 == all_responses
    assert response_pool_2 == all_responses



from starlette_web.tests.core.test_base import BaseTestAPIView


REFERENCE_SCHEMA = {
    "info": {
        "description": "My custom project.",
        "title": "Project documentation",
        "version": "0.0.1"
    },
    "paths": {
        "/api/auth/me/": {
            "get": {
                "description": "Profile info",
                "responses": {
                    "200": {
                        "description": "Profile",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/UserResponse"
                                }
                            }
                        }
                    },
                    "401": {
                        "description": "Authentication error.",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    }
                },
                "tags": [
                    "Authorization"
                ],
                "security": [
                    {
                        "JWTAuth": []
                    }
                ]
            }
        },
        "/api/auth/sign-in/": {
            "post": {
                "description": "Sign in",
                "requestBody": {
                    "required": True,
                    "description": "Sign in",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/SignIn"
                            }
                        }
                    }
                },
                "responses": {
                    "200": {
                        "description": "JsonWebToken",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/JWTResponse"
                                }
                            }
                        }
                    },
                    "401": {
                        "description": "Authentication failed",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    },
                    "400": {
                        "description": "Bad Request.",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    }
                },
                "tags": [
                    "Authorization"
                ]
            }
        },
        "/api/auth/sign-up/": {
            "post": {
                "description": "Sign up",
                "requestBody": {
                    "required": True,
                    "description": "Sign up",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/SignUp"
                            }
                        }
                    }
                },
                "responses": {
                    "200": {
                        "description": "JsonWebToken",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/JWTResponse"
                                }
                            }
                        }
                    },
                    "400": {
                        "description": "Bad Request.",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    }
                },
                "tags": [
                    "Authorization"
                ]
            }
        },
        "/api/auth/sign-out/": {
            "delete": {
                "description": "Sign out",
                "responses": {
                    "200": {
                        "description": "Signed out"
                    },
                    "401": {
                        "description": "Authentication error.",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    }
                },
                "tags": [
                    "Authorization"
                ],
                "security": [
                    {
                        "JWTAuth": []
                    }
                ]
            }
        },
        "/api/auth/refresh-token/": {
            "post": {
                "description": "Update refresh token",
                "requestBody": {
                    "required": True,
                    "description": "Refresh token",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/RefreshToken"
                            }
                        }
                    }
                },
                "responses": {
                    "200": {
                        "description": "JsonWebToken",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/JWTResponse"
                                }
                            }
                        }
                    },
                    "401": {
                        "description": "Authentication failed",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    },
                    "400": {
                        "description": "Bad Request.",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    }
                },
                "tags": [
                    "Authorization"
                ]
            }
        },
        "/api/auth/invite-user/": {
            "post": {
                "description": "Invite user",
                "requestBody": {
                    "required": True,
                    "description": "Invited user's credentials",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/UserInviteRequest"
                            }
                        }
                    }
                },
                "responses": {
                    "200": {
                        "description": "Invited user",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/UserInviteResponse"
                                }
                            }
                        }
                    },
                    "401": {
                        "description": "Authentication error.",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    },
                    "400": {
                        "description": "Bad Request.",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    }
                },
                "tags": [
                    "Authorization"
                ],
                "security": [
                    {
                        "JWTAuth": []
                    }
                ]
            }
        },
        "/api/auth/reset-password/": {
            "post": {
                "description": "Reset password",
                "requestBody": {
                    "required": True,
                    "description": "User credentials",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/ResetPasswordRequest"
                            }
                        }
                    }
                },
                "responses": {
                    "200": {
                        "description": "Link to email",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/ResetPasswordResponse"
                                }
                            }
                        }
                    },
                    "401": {
                        "description": "Authentication error.",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    },
                    "403": {
                        "description": "Access forbidden.",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    },
                    "400": {
                        "description": "Bad Request.",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    }
                },
                "tags": [
                    "Authorization"
                ],
                "security": [
                    {
                        "JWTAuth": []
                    }
                ]
            }
        },
        "/api/auth/change-password/": {
            "post": {
                "description": "Change password",
                "requestBody": {
                    "required": True,
                    "description": "New password confirmation",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/ChangePassword"
                            }
                        }
                    }
                },
                "responses": {
                    "200": {
                        "description": "New tokens",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/JWTResponse"
                                }
                            }
                        }
                    },
                    "401": {
                        "description": "Authentication failed"
                    },
                    "400": {
                        "description": "Bad Request.",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/Error"
                                }
                            }
                        }
                    }
                },
                "tags": [
                    "Authorization"
                ]
            }
        },
        "/health_check/": {
            "get": {
                "description": "Health check of services",
                "responses": {
                    "200": {
                        "description": "Services with status",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HealthCheck"
                                }
                            }
                        }
                    },
                    "503": {
                        "description": "Service unavailable"
                    }
                },
                "tags": [
                    "Health check"
                ]
            }
        }
    },
    "openapi": "3.0.2",
    "components": {
        "schemas": {
            "Error": {
                "type": "object",
                "properties": {
                    "error": {
                        "type": "string"
                    },
                    "details": {
                        "nullable": True
                    }
                },
                "required": [
                    "error"
                ]
            },
            "UserResponse": {
                "type": "object",
                "properties": {
                    "email": {
                        "type": "string",
                        "format": "email"
                    },
                    "id": {
                        "type": "integer"
                    },
                    "is_active": {
                        "type": "boolean"
                    },
                    "is_superuser": {
                        "type": "boolean"
                    }
                },
                "required": [
                    "email",
                    "id",
                    "is_active",
                    "is_superuser"
                ]
            },
            "SignIn": {
                "type": "object",
                "properties": {
                    "email": {
                        "type": "string",
                        "format": "email"
                    },
                    "password": {
                        "type": "string",
                        "minLength": 2,
                        "maxLength": 32
                    }
                },
                "required": [
                    "email",
                    "password"
                ]
            },
            "JWTResponse": {
                "type": "object",
                "properties": {
                    "refresh_token": {
                        "type": "string"
                    },
                    "access_token": {
                        "type": "string"
                    }
                },
                "required": [
                    "access_token",
                    "refresh_token"
                ]
            },
            "SignUp": {
                "type": "object",
                "properties": {
                    "password_2": {
                        "type": "string",
                        "minLength": 2,
                        "maxLength": 32
                    },
                    "email": {
                        "type": "string",
                        "format": "email",
                        "maxLength": 128
                    },
                    "password_1": {
                        "type": "string",
                        "minLength": 2,
                        "maxLength": 32
                    },
                    "invite_token": {
                        "type": "string",
                        "minLength": 10,
                        "maxLength": 32
                    }
                },
                "required": [
                    "email",
                    "invite_token",
                    "password_1",
                    "password_2"
                ]
            },
            "RefreshToken": {
                "type": "object",
                "properties": {
                    "refresh_token": {
                        "type": "string",
                        "minLength": 10,
                        "maxLength": 512
                    }
                },
                "required": [
                    "refresh_token"
                ]
            },
            "UserInviteRequest": {
                "type": "object",
                "properties": {
                    "email": {
                        "type": "string",
                        "format": "email"
                    }
                },
                "required": [
                    "email"
                ]
            },
            "UserInviteResponse": {
                "type": "object",
                "properties": {
                    "token": {
                        "type": "string"
                    },
                    "email": {
                        "type": "string",
                        "format": "email"
                    },
                    "created_at": {
                        "type": "string",
                        "format": "date-time"
                    },
                    "expired_at": {
                        "type": "string",
                        "format": "date-time"
                    },
                    "owner_id": {
                        "type": "integer"
                    },
                    "id": {
                        "type": "integer"
                    }
                },
                "required": [
                    "created_at",
                    "email",
                    "expired_at",
                    "owner_id",
                    "token"
                ]
            },
            "ResetPasswordRequest": {
                "type": "object",
                "properties": {
                    "email": {
                        "type": "string",
                        "format": "email"
                    }
                },
                "required": [
                    "email"
                ]
            },
            "ResetPasswordResponse": {
                "type": "object",
                "properties": {
                    "token": {
                        "type": "string"
                    },
                    "email": {
                        "type": "string",
                        "format": "email"
                    },
                    "user_id": {
                        "type": "integer"
                    }
                },
                "required": [
                    "email",
                    "token"
                ]
            },
            "ChangePassword": {
                "type": "object",
                "properties": {
                    "password_2": {
                        "type": "string",
                        "minLength": 2,
                        "maxLength": 32
                    },
                    "token": {
                        "type": "string",
                        "minLength": 1
                    },
                    "password_1": {
                        "type": "string",
                        "minLength": 2,
                        "maxLength": 32
                    }
                },
                "required": [
                    "password_1",
                    "password_2",
                    "token"
                ]
            },
            "ServicesCheck": {
                "type": "object",
                "properties": {
                    "postgres": {
                        "type": "string"
                    }
                }
            },
            "HealthCheck": {
                "type": "object",
                "properties": {
                    "services": {
                        "$ref": "#/components/schemas/ServicesCheck"
                    },
                    "errors": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    }
                }
            }
        },
        "securitySchemes": {
            "JWTAuth": {
                "type": "http",
                "scheme": "bearer",
                "bearerFormat": "JWT"
            }
        }
    }
}


class TestOpenAPISchema(BaseTestAPIView):
    url = "/openapi"

    def test_schema(self, client):
        response = client.get(self.url)
        assert response.status_code == 200
        assert response.json() == REFERENCE_SCHEMA

    def test_redoc(self, client):
        response = client.get(self.url + "?format=redoc")
        assert response.status_code == 200
        assert b'<script src="/static/apispec/redoc.js"></script>' in response.content



import pytest
from sqlalchemy.ext.asyncio import AsyncSession
from starlette.requests import Request

from starlette_web.common.http.exceptions import (
    AuthenticationRequiredError,
    AuthenticationFailedError,
    PermissionDeniedError,
)
from starlette_web.contrib.auth.backend import JWTAuthenticationBackend
from starlette_web.contrib.auth.models import User, UserSession
from starlette_web.contrib.auth.permissions import IsSuperuserPermission
from starlette_web.contrib.auth.utils import (
    encode_jwt,
    TOKEN_TYPE_RESET_PASSWORD,
    TOKEN_TYPE_REFRESH,
)
from starlette_web.tests.helpers import await_


class TestBackendAuth:
    @staticmethod
    def _prepare_request(db_session: AsyncSession, user: User, user_session: UserSession):
        jwt, _ = encode_jwt({"user_id": user.id, "session_id": user_session.public_id})
        scope = {"type": "http", "headers": [(b"authorization", f"Bearer {jwt}".encode("latin-1"))]}
        request = Request(scope)
        request.state.db_session = db_session
        return request, scope

    def test_check_auth__ok(self, client, user, user_session, dbs):
        request, scope = self._prepare_request(dbs, user, user_session)
        authenticated_user = await_(JWTAuthenticationBackend(request, scope).authenticate())
        assert authenticated_user.id == user.id

    @pytest.mark.parametrize(
        "auth_header, auth_exception, err_details",
        [
            (
                (b"auth", "JWT"),
                AuthenticationRequiredError,
                "Invalid token header. No credentials provided.",
            ),
            (
                (b"authorization", b"JWT"),
                AuthenticationFailedError,
                "Invalid token header. Token should be format as JWT.",
            ),
            (
                (b"authorization", b"FakeKeyword JWT"),
                AuthenticationFailedError,
                "Invalid token header. Keyword mismatch.",
            ),
            (
                (b"authorization", b"Bearer fake-jwt-token"),
                AuthenticationFailedError,
                "Token could not be decoded: Not enough segments",
            ),
        ],
    )
    def test_invalid_token__fail(self, client, user, auth_header, auth_exception, err_details, dbs):
        request = Request(scope={"type": "http", "headers": [auth_header]})
        scope = {"type": "http"}
        request.state.db_session = dbs
        with pytest.raises(auth_exception) as err:
            await_(JWTAuthenticationBackend(request, scope).authenticate())

        assert err.value.details == err_details

    def test_check_auth__user_not_active__fail(self, client, user, user_session, dbs):
        await_(user.update(dbs, is_active=False))
        await_(dbs.commit())
        request, scope = self._prepare_request(dbs, user, user_session)
        with pytest.raises(AuthenticationFailedError) as err:
            await_(JWTAuthenticationBackend(request, scope).authenticate())

        assert err.value.details == f"Couldn't found active user with id={user.id}."

    def test_check_auth__session_not_active__fail(self, client, user, user_session, dbs):
        await_(user_session.update(dbs, is_active=False))
        await_(dbs.commit())
        request, scope = self._prepare_request(dbs, user, user_session)
        with pytest.raises(AuthenticationFailedError) as err:
            await_(JWTAuthenticationBackend(request, scope).authenticate())

        assert err.value.details == (
            f"Couldn't found active session: user_id={user.id} | "
            f"session_id='{user_session.public_id}'."
        )

    @pytest.mark.parametrize("token_type", [TOKEN_TYPE_REFRESH, TOKEN_TYPE_RESET_PASSWORD])
    def test_check_auth__token_t_mismatch__fail(self, client, user, user_session, token_type, dbs):
        await_(user_session.update(dbs, is_active=False))
        await_(dbs.commit())
        token, _ = encode_jwt({"user_id": user.id}, token_type=token_type)
        request, scope = self._prepare_request(dbs, user, user_session)
        with pytest.raises(AuthenticationFailedError) as err:
            await_(
                JWTAuthenticationBackend(request, scope).authenticate_user(
                    token, token_type="access"
                )
            )

        assert err.value.details == f"Token type 'access' expected, got '{token_type}' instead."

    def test_check_auth__admin_required__ok(self, client, user, user_session, dbs):
        await_(user.update(dbs, is_superuser=True))
        await_(dbs.commit())
        request, scope = self._prepare_request(dbs, user, user_session)
        authenticated_user = await_(JWTAuthenticationBackend(request, scope).authenticate())
        assert authenticated_user.id == user.id

        is_admin = await_(IsSuperuserPermission().has_permission(request, scope))
        assert is_admin

    def test_check_auth__admin_required__not_superuser__fail(self, client, user, user_session, dbs):
        await_(user.update(dbs, is_superuser=False))
        await_(dbs.commit())
        request, scope = self._prepare_request(dbs, user, user_session)
        with pytest.raises(PermissionDeniedError) as err:
            await_(JWTAuthenticationBackend(request, scope).authenticate())
            await_(IsSuperuserPermission().has_permission(request, scope))

        assert err.value.details == "You don't have an admin privileges."



import anyio

from starlette_web.common.conf import settings
from starlette_web.common.caches import caches
from starlette_web.common.channels.base import Channel
from starlette_web.common.management.base import BaseCommand
from starlette_web.contrib.redis.channel_layers import RedisPubSubChannelLayer


class Command(BaseCommand):
    help = "Command to cross-process test channels publisher"

    def add_arguments(self, parser):
        parser.add_argument("--group", type=str, required=True)

    async def handle(self, **options):
        group = options["group"]

        redis_options = settings.CHANNEL_LAYERS["redispubsub"]["OPTIONS"]
        channel_ctx = Channel(RedisPubSubChannelLayer(**redis_options))

        with anyio.fail_after(3):
            # Small delay to let subscribers initiate
            await anyio.sleep(0.2)

            async with channel_ctx as channel:
                await channel.publish(group, "Message 0")
                await channel.publish(group, "Message 1")
                await channel.publish(group, "Message 2")
                await channel.publish(group, "DONE")

            await caches["default"].async_set(f"{group}-publisher-done", 1, 10)



from starlette_web.common.utils.serializers import JSONSerializer, PickleSerializer


def test_json_serializer():
    serializer = JSONSerializer()

    obj = {
        "list_1": [{"bool_key": True, "int_key": 1}],
        "obj_key": {"null_key": None, "str_key": "string"},
    }

    encoded = serializer.serialize(obj)
    assert type(encoded) == str

    decoded = serializer.deserialize(encoded)
    assert obj == decoded

    obj = None
    assert serializer.deserialize(serializer.serialize(obj)) is None


def test_pickle_serializer():
    serializer = PickleSerializer()

    obj = {
        "list_1": [{"bool_key": True, "int_key": 1}],
        "obj_key": {"null_key": None, "str_key": "string"},
        "class_key": JSONSerializer,
        "object_key": JSONSerializer(),
        "method_key": JSONSerializer.serialize,
    }

    encoded = serializer.serialize(obj)
    assert type(encoded) == bytes

    decoded = serializer.deserialize(encoded)
    assert obj == decoded



from typing import Any


class Event:
    def __init__(self, group: str, message: Any) -> None:
        self.group = group
        self.message = message

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Event) and self.group == other.group and self.message == other.message
        )

    def __repr__(self) -> str:
        return f"Event(group={self.group!r}, message={self.message!r})"



# flake8: noqa

from starlette_web.contrib.admin import AdminView, admin


"""
Initialize your AdminView's in this file. 
It will be automatically introspected, 
if you add "starlette_web.contrib.admin" and this app's name to settings.INSTALLED_APPS

>>> class ModelView(AdminView):
>>>     pass
>>>
>>> admin.add_view(ModelView(Model, icon="fa fa-users"))
"""
